#!/usr/bin/env python

# Copyright (c) 2009, Giampaolo Rodola'. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""OSX platform implementation."""

import errno
import os
import sys
import warnings

import _psutil_osx
import _psutil_posix
from psutil import _psposix
from psutil._error import AccessDenied, NoSuchProcess, TimeoutExpired
from psutil._compat import namedtuple, wraps
from psutil._common import *

__extra__all__ = []

# --- constants

# Since these constants get determined at import time we do not want to
# crash immediately; instead we'll set them to None and most likely
# we'll crash later as they're used for determining process CPU stats
# and creation_time
try:
    NUM_CPUS = _psutil_osx.get_num_cpus()
except Exception:
    NUM_CPUS = None
    warnings.warn("couldn't determine platform's NUM_CPUS", RuntimeWarning)
try:
    BOOT_TIME = _psutil_osx.get_system_boot_time()
except Exception:
    BOOT_TIME = None
    warnings.warn("couldn't determine platform's BOOT_TIME", RuntimeWarning)
try:
    TOTAL_PHYMEM = _psutil_osx.get_virtual_mem()[0]
except Exception:
    TOTAL_PHYMEM = None
    warnings.warn("couldn't determine platform's TOTAL_PHYMEM", RuntimeWarning)

_PAGESIZE = os.sysconf("SC_PAGE_SIZE")
_cputimes_ntuple = namedtuple('cputimes', 'user nice system idle')
# http://students.mimuw.edu.pl/lxr/source/include/net/tcp_states.h
_TCP_STATES_TABLE = {_psutil_osx.TCPS_ESTABLISHED : CONN_ESTABLISHED,
                     _psutil_osx.TCPS_SYN_SENT : CONN_SYN_SENT,
                     _psutil_osx.TCPS_SYN_RECEIVED : CONN_SYN_RECV,
                     _psutil_osx.TCPS_FIN_WAIT_1 : CONN_FIN_WAIT1,
                     _psutil_osx.TCPS_FIN_WAIT_2 : CONN_FIN_WAIT2,
                     _psutil_osx.TCPS_TIME_WAIT : CONN_TIME_WAIT,
                     _psutil_osx.TCPS_CLOSED : CONN_CLOSE,
                     _psutil_osx.TCPS_CLOSE_WAIT : CONN_CLOSE_WAIT,
                     _psutil_osx.TCPS_LAST_ACK : CONN_LAST_ACK,
                     _psutil_osx.TCPS_LISTEN : CONN_LISTEN,
                     _psutil_osx.TCPS_CLOSING : CONN_CLOSING,
                     _psutil_osx.PSUTIL_CONN_NONE : CONN_NONE,
                     }

# --- functions

get_system_boot_time = _psutil_osx.get_system_boot_time

nt_virtmem_info = namedtuple('vmem', ' '.join([
    # all platforms
    'total', 'available', 'percent', 'used', 'free',
    # OSX specific
    'active',
    'inactive',
    'wired']))

def virtual_memory():
    """System virtual memory as a namedtuple."""
    total, active, inactive, wired, free = _psutil_osx.get_virtual_mem()
    avail = inactive + free
    used = active + inactive + wired
    percent = usage_percent((total - avail), total, _round=1)
    return nt_virtmem_info(total, avail, percent, used, free,
                           active, inactive, wired)

def swap_memory():
    """Swap system memory as a (total, used, free, sin, sout) tuple."""
    total, used, free, sin, sout = _psutil_osx.get_swap_mem()
    percent = usage_percent(used, total, _round=1)
    return nt_swapmeminfo(total, used, free, percent, sin, sout)

def get_system_cpu_times():
    """Return system CPU times as a namedtuple."""
    user, nice, system, idle = _psutil_osx.get_system_cpu_times()
    return _cputimes_ntuple(user, nice, system, idle)

def get_system_per_cpu_times():
    """Return system CPU times as a named tuple"""
    ret = []
    for cpu_t in _psutil_osx.get_system_per_cpu_times():
        user, nice, system, idle = cpu_t
        item = _cputimes_ntuple(user, nice, system, idle)
        ret.append(item)
    return ret

def disk_partitions(all=False):
    retlist = []
    partitions = _psutil_osx.get_disk_partitions()
    for partition in partitions:
        device, mountpoint, fstype, opts = partition
        if device == 'none':
            device = ''
        if not all:
            if not os.path.isabs(device) \
            or not os.path.exists(device):
                continue
        ntuple = nt_partition(device, mountpoint, fstype, opts)
        retlist.append(ntuple)
    return retlist

def get_system_users():
    retlist = []
    rawlist = _psutil_osx.get_system_users()
    for item in rawlist:
        user, tty, hostname, tstamp = item
        if tty == '~':
            continue  # reboot or shutdown
        if not tstamp:
            continue
        nt = nt_user(user, tty or None, hostname or None, tstamp)
        retlist.append(nt)
    return retlist


get_pid_list = _psutil_osx.get_pid_list
pid_exists = _psposix.pid_exists
get_disk_usage = _psposix.get_disk_usage
net_io_counters = _psutil_osx.get_net_io_counters
disk_io_counters = _psutil_osx.get_disk_io_counters

# --- decorator

def wrap_exceptions(fun):
    """Decorator which translates bare OSError exceptions into
    NoSuchProcess and AccessDenied.
    """
    @wraps(fun)
    def wrapper(self, *args, **kwargs):
        try:
            return fun(self, *args, **kwargs)
        except OSError:
            err = sys.exc_info()[1]
            if err.errno == errno.ESRCH:
                raise NoSuchProcess(self.pid, self._process_name)
            if err.errno in (errno.EPERM, errno.EACCES):
                raise AccessDenied(self.pid, self._process_name)
            raise
    return wrapper


_status_map = {
    _psutil_osx.SIDL : STATUS_IDLE,
    _psutil_osx.SRUN : STATUS_RUNNING,
    _psutil_osx.SSLEEP : STATUS_SLEEPING,
    _psutil_osx.SSTOP : STATUS_STOPPED,
    _psutil_osx.SZOMB : STATUS_ZOMBIE,
}

class Process(object):
    """Wrapper class around underlying C implementation."""

    __slots__ = ["pid", "_process_name"]

    def __init__(self, pid):
        self.pid = pid
        self._process_name = None

    @wrap_exceptions
    def get_process_name(self):
        """Return process name as a string of limited len (15)."""
        return _psutil_osx.get_process_name(self.pid)

    @wrap_exceptions
    def get_process_exe(self):
        return _psutil_osx.get_process_exe(self.pid)

    @wrap_exceptions
    def get_process_cmdline(self):
        """Return process cmdline as a list of arguments."""
        if not pid_exists(self.pid):
            raise NoSuchProcess(self.pid, self._process_name)
        return _psutil_osx.get_process_cmdline(self.pid)

    @wrap_exceptions
    def get_process_ppid(self):
        """Return process parent pid."""
        return _psutil_osx.get_process_ppid(self.pid)

    @wrap_exceptions
    def get_process_cwd(self):
        return _psutil_osx.get_process_cwd(self.pid)

    @wrap_exceptions
    def get_process_uids(self):
        real, effective, saved = _psutil_osx.get_process_uids(self.pid)
        return nt_uids(real, effective, saved)

    @wrap_exceptions
    def get_process_gids(self):
        real, effective, saved = _psutil_osx.get_process_gids(self.pid)
        return nt_gids(real, effective, saved)

    @wrap_exceptions
    def get_process_terminal(self):
        tty_nr = _psutil_osx.get_process_tty_nr(self.pid)
        tmap = _psposix._get_terminal_map()
        try:
            return tmap[tty_nr]
        except KeyError:
            return None

    @wrap_exceptions
    def get_memory_info(self):
        """Return a tuple with the process' RSS and VMS size."""
        rss, vms = _psutil_osx.get_process_memory_info(self.pid)[:2]
        return nt_meminfo(rss, vms)

    _nt_ext_mem = namedtuple('meminfo', 'rss vms pfaults pageins')

    @wrap_exceptions
    def get_ext_memory_info(self):
        """Return a tuple with the process' RSS and VMS size."""
        rss, vms, pfaults, pageins = _psutil_osx.get_process_memory_info(self.pid)
        return self._nt_ext_mem(rss, vms,
                                pfaults * _PAGESIZE,
                                pageins * _PAGESIZE)

    @wrap_exceptions
    def get_cpu_times(self):
        user, system = _psutil_osx.get_process_cpu_times(self.pid)
        return nt_cputimes(user, system)

    @wrap_exceptions
    def get_process_create_time(self):
        """Return the start time of the process as a number of seconds since
        the epoch."""
        return _psutil_osx.get_process_create_time(self.pid)

    @wrap_exceptions
    def get_num_ctx_switches(self):
        return nt_ctxsw(*_psutil_osx.get_process_num_ctx_switches(self.pid))

    @wrap_exceptions
    def get_process_num_threads(self):
        """Return the number of threads belonging to the process."""
        return _psutil_osx.get_process_num_threads(self.pid)

    @wrap_exceptions
    def get_open_files(self):
        """Return files opened by process."""
        if self.pid == 0:
            return []
        files = []
        rawlist = _psutil_osx.get_process_open_files(self.pid)
        for path, fd in rawlist:
            if isfile_strict(path):
                ntuple = nt_openfile(path, fd)
                files.append(ntuple)
        return files

    @wrap_exceptions
    def get_connections(self, kind='inet'):
        """Return etwork connections opened by a process as a list of
        namedtuples.
        """
        if kind not in conn_tmap:
            raise ValueError("invalid %r kind argument; choose between %s"
                             % (kind, ', '.join([repr(x) for x in conn_tmap])))
        families, types = conn_tmap[kind]
        rawlist = _psutil_osx.get_process_connections(self.pid, families, types)
        ret = []
        for item in rawlist:
            fd, fam, type, laddr, raddr, status = item
            status = _TCP_STATES_TABLE[status]
            nt = nt_connection(fd, fam, type, laddr, raddr, status)
            ret.append(nt)
        return ret

    @wrap_exceptions
    def get_num_fds(self):
        if self.pid == 0:
            return 0
        return _psutil_osx.get_process_num_fds(self.pid)

    @wrap_exceptions
    def process_wait(self, timeout=None):
        try:
            return _psposix.wait_pid(self.pid, timeout)
        except TimeoutExpired:
            raise TimeoutExpired(self.pid, self._process_name)

    @wrap_exceptions
    def get_process_nice(self):
        return _psutil_posix.getpriority(self.pid)

    @wrap_exceptions
    def set_process_nice(self, value):
        return _psutil_posix.setpriority(self.pid, value)

    @wrap_exceptions
    def get_process_status(self):
        code = _psutil_osx.get_process_status(self.pid)
        if code in _status_map:
            return _status_map[code]
        return constant(-1, "?")

    @wrap_exceptions
    def get_process_threads(self):
        """Return the number of threads belonging to the process."""
        rawlist = _psutil_osx.get_process_threads(self.pid)
        retlist = []
        for thread_id, utime, stime in rawlist:
            ntuple = nt_thread(thread_id, utime, stime)
            retlist.append(ntuple)
        return retlist

    nt_mmap_grouped = namedtuple('mmap',
        'path rss private swapped dirtied ref_count shadow_depth')
    nt_mmap_ext = namedtuple('mmap',
        'addr perms path rss private swapped dirtied ref_count shadow_depth')

    @wrap_exceptions
    def get_memory_maps(self):
        return _psutil_osx.get_process_memory_maps(self.pid)



# mt_histgmm.py
# KMB Matlab
# translated to Python 4/12/07 JAB

from matplotlib.mlab import normpdf
import numpy as num
from scipy.linalg.basic import eps, kron # kron substitutes for repmat
import sys

PRINT_PROGRESS = False

def make_vect( arr ):
    return arr.reshape( arr.size )

def histgmm( x, w, maxiters=100, minchange=None, minsigma=None ):
    """Gaussian mixture modeling using 2 clusters.
    Returns 2-D arrays 'mu', 'sigma', 'prior'."""

    if len(w.shape) == 1:
        n = w.size
        m = 1
        w = w.reshape( (n,m) )
        x = x.reshape( (n,m) )
    else:
        n,m = w.shape
    w /= w.sum( 0 )
    x = kron( num.ones((1,m)), x )

    if minchange is None:
        minchange = (x[-1]-x[0]) / n/2
    if minsigma is None:
        minsigma = (x[-1]-x[0]) / n/2

    ## initialize ##

    # first initialization scheme
    mu0 = num.zeros( (2,m) )
    mu0[0,:] = x[0,:]
    mu0[1,:] = x[-1,:]
    sigma0 = (x[-1]-x[0]) / 4 * num.ones( (2,m) )
    gamma = 0.5 * num.ones( (n,m) )
    prior = 0.5 * num.ones( (2,m) )

    mu, sigma, prior, score1 = histgmm_main( mu0, sigma0, gamma, prior, w, x,
                                             maxiters, minchange, minsigma, n, m )

    # second initialization scheme
    c = num.r_[num.zeros(m), w.cumsum()]
    gamma = (c[:-1] < 0.5).astype( num.float )
    gamma.resize( (n,m) )
    mu0 = num.zeros( (2,m) )
    sigma0 = num.zeros( (2,m) )
    tmp = w * gamma
    prior[0,:] = tmp.sum( 0 )
    isprior = prior[0,:] > 0
    # the 'if' statement is necessary because sum( [] ) returns 0, not empty
    if isprior.any():
        mu0[0,isprior] = (tmp * x[:,isprior]).sum(0) / prior[0,isprior]
        a = (x[:,isprior].T - \
              kron( num.ones((n,1)), mu0[0,isprior] )).T
        sigma0[0,isprior] = num.sqrt( (tmp * a**2).sum(0) / prior[0,isprior] )
    tmp = w * (1.-gamma)
    prior[1,:] = tmp.sum( 0 )
    isprior = prior[1,:] > 0
    if isprior.any():
        mu0[1,isprior] = (tmp * x[:,isprior]).sum(0) / prior[1,isprior]
        a = (x[:,isprior].T - \
              kron( num.ones((n,1)), mu0[1,isprior] )).T
        sigma0[1,isprior] = num.sqrt( (tmp * a**2).sum(0) / prior[1,isprior] )
    sigma0[sigma0 < minsigma] = minsigma

    mu2, sigma2, prior2, score2 = histgmm_main( mu0, sigma0, gamma, prior, w, x,
                                                maxiters, minchange, minsigma, n, m )

    ## choose best of mu, mu2
    
    # vectorize scores to allow indexing
    inds2 = num.array((score2,0.)) > num.array((score1,1.))
    if inds2.any():
        mu[:,inds2] = mu2[:,inds2]
        sigma[:,inds2] = sigma2[:,inds2]
        prior[:,inds2] = prior2[:,inds2]

    noprior = prior[0,:] < eps
    if noprior.any():
        mu[0,noprior] = mu[1,noprior]
        sigma[0,noprior] = sigma[1,noprior]
    noprior = prior[1,:] < eps
    if noprior.any():
        mu[1,noprior] = mu[0,noprior]
        sigma[1,noprior] = sigma[0,noprior]

    return mu, sigma, prior

def histgmm_main( mu0, sigma0, gamma, prior, w, x,
                  maxiters, minchange, minsigma, n, m ):
    """Do EM."""
    mu = num.zeros( mu0.shape )
    sigma = num.zeros( mu0.shape )

    ischange = num.ones( m ).astype( num.bool )

    if PRINT_PROGRESS: print "EM iterations:", ; sys.stdout.flush()
    for iter in range( maxiters ):
        if PRINT_PROGRESS: print iter, ; sys.stdout.flush()

        # compute probability for each
        p1 = normpdf( make_vect( x[:,ischange] ).T,
                      make_vect( kron( num.ones((n,1)), mu0[0,ischange] ) ).T,
                      make_vect( kron( num.ones((n,1)), sigma0[0,ischange] ) ) ) \
                      * make_vect( kron( num.ones((n,1)), prior[0,ischange] ) )
        p2 = normpdf( make_vect( x[:,ischange] ).T,
                      make_vect( kron( num.ones((n,1)), mu0[1,ischange] ) ).T,
                      make_vect( kron( num.ones((n,1)), sigma0[1,ischange] ) ) ) \
                      * make_vect( kron( num.ones((n,1)), prior[1,ischange] ) )
        tmp = p1 + p2
        p1[tmp < eps] = 0.5
        p2[tmp < eps] = 0.5
        tmp[tmp < eps] = 1
        gamma[:,ischange] = (p1/tmp).reshape( (n,len(ischange.nonzero()[0])) )

        # update the mean, variance, and prior
        tmp = w[:,ischange] * gamma[:,ischange]
        prior[0,ischange] = tmp.sum( 0 )
        tmp = tmp[:,prior[0,ischange] > 0]
        isprior = prior[0,:] > 0
        c_and_p = num.logical_and( ischange, isprior )
        if c_and_p.any():
            mu[0,c_and_p] = (tmp * x[:,c_and_p]).sum(0) / prior[0,c_and_p]
            a = (x[:,c_and_p].T - \
                  kron( num.ones((n,1)), mu[0,c_and_p] )).T
            sigma[0,c_and_p] = num.sqrt( (tmp * a**2).sum(0) / prior[0,c_and_p] )
        
        tmp = w[:,ischange] * (1-gamma[:,ischange])
        prior[1,ischange] = tmp.sum( 0 )
        tmp = tmp[:,prior[1,ischange] > 0]
        isprior = prior[1,:] > 0
        c_and_p = num.logical_and( ischange, isprior )
        if c_and_p.any():
            mu[1,c_and_p] = (tmp * x[:,c_and_p]).sum(0) / prior[1,c_and_p]
            a = (x[:,c_and_p].T - \
                  kron( num.ones((n,1)), mu[1,c_and_p] )).T
            sigma[1,c_and_p] = num.sqrt( (tmp * a**2).sum(0) / prior[1,c_and_p] )

        sigma[sigma < minsigma] = minsigma

        # see if there is a change
        ischange = num.logical_or( (num.abs( mu - mu0 ) >= minchange),
                                   (num.abs( sigma - sigma0 ) >= minchange) ).any(0)
        if not ischange.any(): break

        mu0 = mu.copy()
        sigma0 = sigma.copy()
    # endfor: EM iterations
    if PRINT_PROGRESS: print

    p1 = normpdf( make_vect( x ).T,
                  make_vect( kron( num.ones((n,1)), mu[0,:] ) ).T,
                  make_vect( kron( num.ones((n,1)), sigma[0,:] ) ) ) \
                  * make_vect( kron( num.ones((n,1)), prior[0,:] ) )
    p2 = normpdf( make_vect( x ).T,
                  make_vect( kron( num.ones((n,1)), mu[1,:] ) ).T,
                  make_vect( kron( num.ones((n,1)), sigma[1,:] ) ) ) \
                  * make_vect( kron( num.ones((n,1)), prior[1,:] ) )
    p1.reshape( (n,m) )
    p2.reshape( (n,m) )
    a = make_vect(w) * (make_vect(gamma)*p1 + (1-make_vect(gamma))*p2)
    score = a.sum( 0 )

    return mu, sigma, prior, score



import numpy as num
import scipy.stats as stats
from params import params
import annfiles as annot
import matplotlib.pyplot as plt
import os.path
import pickle
import getopt
import sys

if params.interactive:
    import wx

DEBUG = True
DELTAPLOT = .05

class ExpTrackingSettings:

    def __init__( self, annnames=None,
                  picklefile=None, 
                  nframessample = 100,
                  min_major_percentile = 1.,
                  max_major_percentile = 99.,
                  min_minor_percentile = 1.,
                  max_minor_percentile = 99.,
                  min_area_percentile = 1.,
                  max_area_percentile = 99.,
                  min_ecc_percentile = 1.,
                  max_ecc_percentile = 99.,
                  jump_distance_delta = .1
                  ):
        
        if picklefile is not None:
            if params.interactive:
                wx.Yield()
                wx.BeginBusyCursor()
            else:
                print 'Loading pickled tracking settings models'
            fid = open(picklefile,'r')
            model = pickle.load(fid)
            fid.close()

            for key,val in model.iteritems():
                setattr(self,key,val)

            if params.interactive:
                wx.EndBusyCursor()
        else:
        
            self.annnames = annnames
            self.nmovies = len(annnames)
    
            self.nframessample = nframessample
            self.min_major_percentile = min_major_percentile
            self.max_major_percentile = max_major_percentile
            self.min_minor_percentile = min_minor_percentile
            self.max_minor_percentile = max_minor_percentile
            self.min_area_percentile = min_area_percentile
            self.max_area_percentile = max_area_percentile
            self.min_ecc_percentile = min_ecc_percentile
            self.max_ecc_percentile = max_ecc_percentile
            self.jump_distance_delta = jump_distance_delta
            
            # initialize models
            self.init_shape_models()
            self.init_motion_models()

    def init_shape_models(self):

        self.majors = []
        self.minors = []
        self.areas = []
        self.eccs = []
        self.movieis = []
        self.flys = []
        self.frames = []

    def init_motion_models(self):

        self.xcurrs = []
        self.ycurrs = []
        self.anglecurrs = []
        self.xprevs = []
        self.yprevs = []
        self.angleprevs = []
        self.dxs = []
        self.dys = []
        self.dangles = []
        self.is_motion_data = []

    def update_shape_models(self):

        for ell in self.trxcurr.itervalues():
            self.majors.append(ell.size.height*4.)
            self.minors.append(ell.size.width*4.)
            self.areas.append(ell.area())
            self.eccs.append(ell.size.width / ell.size.height)
            self.movieis.append(self.moviei)
            self.flys.append(ell.identity)
            self.frames.append(self.frame)

    def update_motion_models(self):

        for id,ellcurr in self.trxcurr.iteritems():
            if self.trxprev.hasItem(id) and self.trxprevprev.hasItem(id):
                ellprev = self.trxprev[id]
                ellprevprev = self.trxprevprev[id]
                dx = ellprev.center.x - ellprevprev.center.x
                dy = ellprev.center.y - ellprevprev.center.y
                dangle = ((ellprev.angle - ellprevprev.angle + num.pi/2.) \
                              % (num.pi)) - (num.pi/2.)
                anglepred = ellprev.angle + dangle
                anglecurr = (ellcurr.angle - anglepred + num.pi/2.) % num.pi + anglepred - num.pi/2
                self.xcurrs.append(ellcurr.center.x)
                self.ycurrs.append(ellcurr.center.y)
                self.xprevs.append(ellprev.center.x)
                self.yprevs.append(ellprev.center.y)
                self.anglecurrs.append(anglecurr)
                self.angleprevs.append(ellprev.angle)
                self.dxs.append(dx)
                self.dys.append(dy)
                self.dangles.append(dangle)
                self.is_motion_data.append(True)

            else:

                self.xcurrs.append(num.nan)
                self.ycurrs.append(num.nan)
                self.xprevs.append(num.nan)
                self.yprevs.append(num.nan)
                self.anglecurrs.append(num.nan)
                self.angleprevs.append(num.nan)
                self.dxs.append(num.nan)
                self.dys.append(num.nan)
                self.dangles.append(num.nan)
                self.is_motion_data.append(False)
    
    def compute_models_permovie( self, i ):
        """
        compute_models_permovie( i )
        
        For annnames[i], this function samples frames from throughout
        the movie and computes the data structures necessary for 
        estimating the tracking settings.
        """

        self.moviei = i
        # open movie
        #self.movie = movies.Movie( self.movienames[i], params.interactive )

        # background model
        #self.bg_imgs = bg.BackgroundCalculator(self.movie)

        # open annotation
        self.trx = annot.AnnotationFile( self.annnames[i], doreadbgmodel=False, 
                                         doreadtrx=True )


        # choose frames to learn from: for now, assume all frames are tracked
        self.firstframe = self.trx.firstframetracked+2
        self.lastframe = self.trx.lastframetracked

        self.framessample = num.unique(num.round(num.linspace(self.firstframe+1,
                                                              self.lastframe,
                                                              self.nframessample)).astype(int))

        if DEBUG: print "Collecting data for movie %d, annfile = %s..."%(self.moviei,self.annnames[self.moviei])

        for j in range(len(self.framessample)):

            # read in the sampled frame
            self.frame = self.framessample[j]
            self.trxcurr = self.trx.get_frame(self.frame)
            self.trxprev = self.trx.get_frame(self.frame-1)
            self.trxprevprev = self.trx.get_frame(self.frame-2)

            if DEBUG: print "Collecting data for frame %d"%self.frame

            # update all the observation models
            self.update_shape_models()
            self.update_motion_models()

        #self.movie.close()
        #self.movie = None
        self.trx.close()
        self.trx = None
        #self.bg_imgs = None
        
    def est_settings(self):
        """
        est_settings()

        """

        for i in range(self.nmovies):

            if DEBUG: print "Computing per-movie model for movie %d"%i
            self.compute_models_permovie(i)

        self.est_shape_parameters()
        self.est_motion_parameters()

    def est_shape_parameters(self):

        self.majors = num.array(self.majors)
        self.minors = num.array(self.minors)
        self.areas = num.array(self.areas)
        self.eccs = num.array(self.eccs)

        self.min_major = genprctile(self.majors,self.min_major_percentile)
        self.max_major = genprctile(self.majors,self.max_major_percentile)
        self.min_minor = genprctile(self.minors,self.min_minor_percentile)
        self.max_minor = genprctile(self.minors,self.max_minor_percentile)
        self.min_area = genprctile(self.areas,self.min_area_percentile)
        self.max_area = genprctile(self.areas,self.max_area_percentile)
        self.min_ecc = genprctile(self.eccs,self.min_ecc_percentile)
        self.max_ecc = genprctile(self.eccs,self.max_ecc_percentile)
        self.mean_major = num.mean(self.majors)
        self.mean_minor = num.mean(self.minors)
        self.mean_area = num.mean(self.areas)
        self.mean_ecc = num.mean(self.eccs)

    def est_motion_parameters(self):

        self.xcurrs = num.array(self.xcurrs)
        self.ycurrs = num.array(self.ycurrs)
        self.anglecurrs = num.array(self.anglecurrs)
        self.xprevs = num.array(self.xprevs)
        self.yprevs = num.array(self.yprevs)
        self.angleprevs = num.array(self.angleprevs)
        self.dxs = num.array(self.dxs)
        self.dys = num.array(self.dys)
        self.dangles = num.array(self.dangles)
        self.is_motion_data = num.array(self.is_motion_data,dtype=bool)

        # min( [ (xcurrs-xprevs) - alpha*dxs ].T * (xcurrs-xprevs) - alpha*dxs ] + \
        #      [ (ycurrs-yprevs) - alpha*dys ].T * (ycurrs-yprevs) - alpha*dys ] )
        # =>
        # ((xcurrs-xprevs) - alpha*dxs).T*dxs + 
        # ((ycurrs-yprevs) - alpha*dxs).T*dxs = 0
        # (xcurrs-xprevs).T * dxs + (ycurrs-yprevs).T * dys = alpha*(dxs.T * dxs + dys.T * dy)
        # alpha = [ (xcurrs-xprevs).T * dxs + (ycurrs-yprevs).T * dys ] / (dxs.T * dxs + dys.T * dy)
        alpha = ( num.sum( (self.xcurrs[self.is_motion_data]-self.xprevs[self.is_motion_data])*self.dxs[self.is_motion_data] ) + \
                  num.sum( (self.ycurrs[self.is_motion_data]-self.yprevs[self.is_motion_data])*self.dys[self.is_motion_data] ) ) / \
                ( num.sum( self.dxs[self.is_motion_data]**2. ) + num.sum( self.dys[self.is_motion_data]**2. ) )
        alpha = max(alpha,0.)
        alpha = min(alpha,1.)
        self.center_dampen = 1. - alpha

        alpha = num.sum( self.anglecurrs[self.is_motion_data]*self.dangles[self.is_motion_data] ) / \
                num.sum( self.dangles[self.is_motion_data]**2. )
        alpha = max(alpha,0.)
        alpha = min(alpha,1.)
        self.angle_dampen = 1. - alpha

        # choose the weight of angle error
        self.xpreds = self.xprevs + self.dxs*(1.-self.center_dampen)
        self.ypreds = self.yprevs + self.dys*(1.-self.center_dampen)
        self.anglepreds = self.angleprevs + self.dangles*(1.-self.angle_dampen)
        self.center_err2s = (self.xpreds-self.xcurrs)**2 + (self.ypreds-self.ycurrs)**2.
        self.center_err2 = num.mean(self.center_err2s[self.is_motion_data])
        self.angle_err2s = ((self.anglepreds-self.anglecurrs+num.pi/2.)%num.pi - num.pi/2.)**2.
        self.angle_err2 = num.mean(self.angle_err2s[self.is_motion_data])
        self.angle_weight = self.center_err2 / self.angle_err2

        # choose the maximum jump distance
        self.dists = num.sqrt(self.center_err2s +self.angle_weight*self.angle_err2s)
        self.max_jump_distance = num.max(self.dists[self.is_motion_data])*(1.+self.jump_distance_delta)

    def save(self,outputFileName):
        
        fid = open(outputFileName,"w")
        
        out = dict()
        if hasattr(self,'nframessample'):
            out['nframessample'] = self.nframessample
        if hasattr(self,'min_major_percentile'):
            out['min_major_percentile'] = self.min_major_percentile
        if hasattr(self,'max_major_percentile'):
            out['max_major_percentile'] = self.max_major_percentile
        if hasattr(self,'min_minor_percentile'):
            out['min_minor_percentile'] = self.min_minor_percentile
        if hasattr(self,'max_minor_percentile'):
            out['max_minor_percentile'] = self.max_minor_percentile
        if hasattr(self,'min_area_percentile'):
            out['min_area_percentile'] = self.min_area_percentile
        if hasattr(self,'max_area_percentile'):
            out['max_area_percentile'] = self.max_area_percentile
        if hasattr(self,'min_ecc_percentile'):
            out['min_ecc_percentile'] = self.min_ecc_percentile
        if hasattr(self,'max_ecc_percentile'):
            out['max_ecc_percentile'] = self.max_ecc_percentile
        if hasattr(self,'jump_distance_delta'):
            out['jump_distance_delta'] = self.jump_distance_delta
        if hasattr(self,'annnames'):
            out['annnames'] = self.annnames

        if hasattr(self,'majors'):
            out['majors'] = self.majors
        if hasattr(self,'minors'):
            out['minors'] = self.minors
        if hasattr(self,'areas'):
            out['areas'] = self.areas
        if hasattr(self,'eccs'):
            out['eccs'] = self.eccs
        if hasattr(self,'movieis'):
            out['movieis'] = self.movieis
        if hasattr(self,'flys'):
            out['flys'] = self.flys
        if hasattr(self,'frames'):
            out['frames'] = self.frames
        if hasattr(self,'xcurrs'):
            out['xcurrs'] = self.xcurrs
        if hasattr(self,'ycurrs'):
            out['ycurrs'] = self.ycurrs
        if hasattr(self,'anglecurrs'):
            out['anglecurrs'] = self.anglecurrs
        if hasattr(self,'xprevs'):
            out['xprevs'] = self.xprevs
        if hasattr(self,'yprevs'):
            out['yprevs'] = self.yprevs
        if hasattr(self,'angleprevs'):
            out['angleprevs'] = self.angleprevs
        if hasattr(self,'dxs'):
            out['dxs'] = self.dxs
        if hasattr(self,'dys'):
            out['dys'] = self.dys
        if hasattr(self,'dangles'):
            out['dangles'] = self.dangles
        if hasattr(self,'is_motion_data'):
            out['is_motion_data'] = self.is_motion_data
        if hasattr(self,'min_major'):
            out['min_major'] = self.min_major
        if hasattr(self,'max_major'):
            out['max_major'] = self.max_major
        if hasattr(self,'min_minor'):
            out['min_minor'] = self.min_minor
        if hasattr(self,'max_minor'):
            out['max_minor'] = self.max_minor
        if hasattr(self,'min_area'):
            out['min_area'] = self.min_area
        if hasattr(self,'max_area'):
            out['max_area'] = self.max_area
        if hasattr(self,'min_ecc'):
            out['min_ecc'] = self.min_ecc
        if hasattr(self,'max_ecc'):
            out['max_ecc'] = self.max_ecc
        if hasattr(self,'mean_major'):
            out['mean_major'] = self.mean_major
        if hasattr(self,'mean_minor'):
            out['mean_minor'] = self.mean_minor
        if hasattr(self,'mean_area'):
            out['mean_area'] = self.mean_area
        if hasattr(self,'mean_ecc'):
            out['mean_ecc'] = self.mean_ecc
        if hasattr(self,'center_dampen'):
            out['center_dampen'] = self.center_dampen
        if hasattr(self,'angle_dampen'):
            out['angle_dampen'] = self.angle_dampen
        if hasattr(self,'xpreds'):
            out['xpreds'] = self.xpreds
        if hasattr(self,'ypreds'):
            out['ypreds'] = self.ypreds
        if hasattr(self,'anglepreds'):
            out['anglepreds'] = self.anglepreds
        if hasattr(self,'center_err2s'):
            out['center_err2s'] = self.center_err2s
        if hasattr(self,'angle_err2s'):
            out['angle_err2s'] = self.angle_err2s
        if hasattr(self,'center_err2'):
            out['center_err2'] = self.center_err2
        if hasattr(self,'angle_err2'):
            out['angle_err2'] = self.angle_err2
        if hasattr(self,'angle_weight'):
            out['angle_weight'] = self.angle_weight
        if hasattr(self,'dists'):
            out['dists'] = self.dists
        if hasattr(self,'max_jump_distance'):
            out['max_jump_distance'] = self.max_jump_distance

        pickle.dump(out,fid)
        fid.close()

    def show(self):
        
        self.show_shape()
        self.show_motion()
        plt.show()

    def show_motion(self):
        
        nbins = 100

        self.counts_centererr,self.edges_centererr = num.histogram(num.sqrt(self.center_err2s[self.is_motion_data]),nbins)
        self.counts_centererr = self.counts_centererr.astype(float)
        self.frac_centererr = self.counts_centererr / sum(self.counts_centererr)
        self.centers_centererr = (self.edges_centererr[:-1]+self.edges_centererr[1:])/2.

        self.counts_angleerr,self.edges_angleerr = num.histogram(num.sqrt(self.angle_err2s[self.is_motion_data]),nbins)
        self.counts_angleerr = self.counts_angleerr.astype(float)
        self.frac_angleerr = self.counts_angleerr / sum(self.counts_angleerr)
        self.centers_angleerr = (self.edges_angleerr[:-1]+self.edges_angleerr[1:])/2.

        self.counts_dist,self.edges_dist = num.histogram(self.dists[self.is_motion_data],nbins)
        self.counts_dist = self.counts_dist.astype(float)
        self.frac_dist = self.counts_dist / num.sum(self.counts_dist)
        self.centers_dist = (self.edges_dist[:-1]+self.edges_dist[1:])/2.

        plt.figure()

        plt.subplot(131)        
        plt.plot(self.centers_centererr,self.frac_centererr,'k.-')
        maxy = num.max(self.frac_centererr)
        ax = num.array([self.edges_centererr[0],self.edges_centererr[-1],-maxy*DELTAPLOT,maxy*(1.+DELTAPLOT)])
        plt.axis(ax)
        plt.title('center dist')

        plt.subplot(132)        
        plt.plot(self.centers_angleerr,self.frac_angleerr,'k.-')
        maxy = num.max(self.frac_angleerr)
        ax = num.array([self.edges_angleerr[0],self.edges_angleerr[-1],-maxy*DELTAPLOT,maxy*(1.+DELTAPLOT)])
        plt.axis(ax)
        plt.title('angle dist')
        
        plt.subplot(133)        
        plt.plot(self.centers_dist,self.frac_dist,'k.-')
        plt.hold(True)
        maxy = num.max(self.frac_dist)
        ax = num.array([self.edges_dist[0],max(self.max_jump_distance*(1.+DELTAPLOT),self.edges_dist[-1]),-maxy*DELTAPLOT,maxy*(1.+DELTAPLOT)])
        plt.plot(self.max_jump_distance+num.zeros(2),ax[[2,3]],'r-')
        plt.axis(ax)
        plt.title('dist')

        print "angle weight = " + str(self.angle_weight)
        print "center dampen = " + str(self.center_dampen)
        print "angle dampen = " + str(self.angle_dampen)
        print "max jump dist = " + str(self.max_jump_distance)

    def show_shape(self):

        nbins = 100

        # show shape
        
        # histogram of area axis lengths
        self.counts_area,self.edges_area = num.histogram(self.areas,nbins)
        self.counts_area = self.counts_area.astype(float)
        self.frac_area = self.counts_area / num.sum(self.counts_area)
        self.centers_area = (self.edges_area[:-1]+self.edges_area[1:])/2.

        # plot histogram
        plt.figure()

        plt.subplot(221)
        plt.plot(self.centers_area,self.frac_area,'k.-')
        plt.hold(True)
        ax = get_axis(self.edges_area,self.frac_area,self.min_area,self.max_area)
        plt.plot(self.min_area+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.mean_area+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.max_area+num.zeros(2),ax[[2,3]],'r-')
        plt.axis(ax)
        plt.title('area')

        # histogram of major axis lengths
        self.counts_major,self.edges_major = num.histogram(self.majors,nbins)
        self.counts_major = self.counts_major.astype(float)
        self.frac_major = self.counts_major / num.sum(self.counts_major)
        self.centers_major = (self.edges_major[:-1]+self.edges_major[1:])/2.

        # plot histogram
        plt.subplot(222)
        plt.plot(self.centers_major,self.frac_major,'k.-')
        plt.hold(True)
        ax = get_axis(self.edges_major,self.frac_major,self.min_major,self.max_major)
        plt.plot(self.min_major+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.mean_major+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.max_major+num.zeros(2),ax[[2,3]],'r-')
        plt.axis(ax)
        plt.title('major')

        # histogram of minor axis lengths
        self.counts_minor,self.edges_minor = num.histogram(self.minors,nbins)
        self.counts_minor = self.counts_minor.astype(float)
        self.frac_minor = self.counts_minor / num.sum(self.counts_minor)
        self.centers_minor = (self.edges_minor[:-1]+self.edges_minor[1:])/2.

        # plot histogram
        plt.subplot(223)
        plt.plot(self.centers_minor,self.frac_minor,'k.-')
        plt.hold(True)
        ax = get_axis(self.edges_minor,self.frac_minor,self.min_minor,self.max_minor)
        plt.plot(self.min_minor+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.mean_minor+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.max_minor+num.zeros(2),ax[[2,3]],'r-')
        plt.axis(ax)
        plt.title('minor')

        # histogram of ecc axis lengths
        self.counts_ecc,self.edges_ecc = num.histogram(self.eccs,nbins)
        self.counts_ecc = self.counts_ecc.astype(float)
        self.frac_ecc = self.counts_ecc / num.sum(self.counts_ecc)
        self.centers_ecc = (self.edges_ecc[:-1]+self.edges_ecc[1:])/2.

        # plot histogram
        plt.subplot(224)
        plt.plot(self.centers_ecc,self.frac_ecc,'k.-')
        plt.hold(True)
        ax = get_axis(self.edges_ecc,self.frac_ecc,self.min_ecc,self.max_ecc)
        plt.plot(self.min_ecc+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.mean_ecc+num.zeros(2),ax[[2,3]],'r-')
        plt.plot(self.max_ecc+num.zeros(2),ax[[2,3]],'r-')
        plt.axis(ax)
        plt.title('ecc')

        print "min area = " + str(self.min_area)
        print "mean area = " + str(self.mean_area)
        print "max area = " + str(self.max_area)
        print "min major = " + str(self.min_major)
        print "mean major = " + str(self.mean_major)
        print "max major = " + str(self.max_major)
        print "min minor = " + str(self.min_minor)
        print "mean minor = " + str(self.mean_minor)
        print "max minor = " + str(self.max_minor)
        print "min ecc = " + str(self.min_ecc)
        print "mean ecc = " + str(self.mean_ecc)
        print "max ecc = " + str(self.max_ecc)

def main():
    """
    main()

    Compute experiment shape and motion models.
    """

    shortopts = "f:p:a:o:"
    longopts = ["filename=","params=","ann=","output="]
    try:
        opts, args = getopt.getopt(sys.argv[1:], shortopts, longopts)
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        print "ExpTrackingSettings options:"
        print "    -f,--filename <name of file containing list of experiment directories>"
        print "      Default: expdirs.txt"
        print "    -p,--params <name of file containing parameters>"
        print "      Default: <empty string>"
        print "    -a,--ann <name of annotation file within experiment directories>"
        print "      Default: movie.ufmf.ann"
        print "    -o,--output <name of file to output results to>"
        print "      Default: ExpTrackingSettingsResults.pickle"
        sys.exit(2)
        
    expdirsFileName = 'expdirs.txt'
    paramsFileName = ''
    annFileStr = 'movie.ufmf.ann'
    outputFileName = 'ExpTrackingSettingsResults.pickle'
    
    for o,a in opts:
        if o in ("-f","--filename"):
            expdirsFileName = a
            print "expdirsFileName = " + a
        elif o in ("-p","--params"):
            paramsFileName = a
            print "paramsFileName = " + a
        elif o in ("-a","--ann"):
            annFileStr = a
            print "annFileStr = " + a
        elif o in ("-o","--output"):
            outputFileName = a
            print "outputFileName = " + a
        else:
            assert False, "unhandled option"

    # read in the experiment directories
    fid = open(expdirsFileName,"r")
    expdirs = []
    annnames = []
    for l in fid:
        expdir = l.strip()
        if not os.path.exists(expdir):
            print "Experiment directory %s does not exist. Skipping."%expdir
            continue
        annname = os.path.join(expdir,annFileStr)
        if not os.path.exists(annname):
            print "Annotation file %s does not exist. Skipping experiment %s."%(annname,expdir)
            continue
        expdirs.append(expdir)
        annnames.append(annname)
    fid.close()
        
    if paramsFileName != '':
        params = read_params(paramsFileName)
        
    model = ExpTrackingSettings(annnames,
                                nframessample = params['nframessample'],
                                min_major_percentile = params['min_major_percentile'],
                                max_major_percentile = params['max_major_percentile'],
                                min_minor_percentile = params['min_minor_percentile'],
                                max_minor_percentile = params['max_minor_percentile'],
                                min_area_percentile = params['min_area_percentile'],
                                max_area_percentile = params['max_area_percentile'],
                                min_ecc_percentile = params['min_ecc_percentile'],
                                max_ecc_percentile = params['max_ecc_percentile'],
                                jump_distance_delta = params['jump_distance_delta'])
    model.est_settings()
    model.save(outputFileName)
    model.show()

    return model
    
def genprctile(x,prct):
    if prct < 0 or prct > 100:
        maxx = num.max(x)
        minx = num.min(x)
        dx = maxx - minx
        if prct < 0:
            p = minx - (-prct/100.)*dx
        else:
            p = maxx + ((prct-100.)/100)*dx
    else:
        p = stats.scoreatpercentile(x,prct)
    return p

def read_params(paramsFileName):

    # default parameters
    params = dict()
    params['nframessample'] = 100
    params['min_major_percentile'] = 1.
    params['max_major_percentile'] = 99.
    params['min_minor_percentile'] = 1.
    params['max_minor_percentile'] = 99.
    params['min_area_percentile'] = 1.
    params['max_area_percentile'] = 99.
    params['min_ecc_percentile'] = 1.
    params['max_ecc_percentile'] = 99.
    params['jump_distance_delta'] = .1

    # parse parameters file
    fid = open(paramsFileName,"r")

    for l in fid:
        l = l.strip()
        if l[0] == '#':
            continue
        l = l.split('=',1)
        if len(l) < 2:
            print "Skipping parameter line '%s'"%l
            continue
        l[0] = l[0].strip()
        l[1] = l[1].strip()
        if l[0] in params.keys():
            params[l[0]] = float(l[1])
    fid.close()

    params['nframessample'] = int(params['nframessample'])

    return params

def get_axis(edges,frac,minv,maxv):
    ax = num.zeros(4)
    dv = edges[-1] - edges[0]
    maxfrac = num.max(frac)
    ax[0] = min(edges[0],minv-dv*DELTAPLOT)
    ax[1] = max(edges[-1],maxv+dv*DELTAPLOT)
    ax[2] = 0-maxfrac*DELTAPLOT
    ax[3] = maxfrac*(1.+DELTAPLOT)
    return ax


if __name__ == "__main__":
    params.interactive = False
    main()



# movies.py
# KMB 11/06/2008

import chunk
import multiprocessing
import os
import struct
import sys
import traceback

import numpy as num
import wx

from params import params
from ellipsesk import annotate_bmp
from version import USE_AVBIN
if USE_AVBIN:
    import pyglet.media as media

# version of sbfmf for writing
__version__ = "0.3b"

#import static_background_fmf as sbfmf
#import motmot.FlyMovieFormat.FlyMovieFormat as fmf
import FlyMovieFormat as fmf
try:
    from FlyMovieFormat import NoMoreFramesException
except ImportError:
    class NoMoreFramesException (Exception): pass
try:
    # FIXME: change back to motmot.ufmf.ufmf
    #import motmot.ufmf.ufmf as ufmf
    import ufmf
except ImportError:
    pass

DEBUG_MOVIES = True

from version import DEBUG
if not DEBUG:
    DEBUG_MOVIES = False


def known_extensions():
    return ['.fmf', '.avi', '.sbfmf', '.ufmf'] # must sync with line 75


class Movie:
    """Generic interface for all supported movie types."""
    def __init__( self, initpath,
                  interactive=True,
                  parentframe=None,
                  open_now=True,
                  default_extension='.fmf' ):
        """Prepare to open a movie (awaiting call to self.open()).
If initpath is a filename, just use it.
If initpath is a directory and interactive is True, then ask user for a filename.
If initpath is a directory and not in interactive mode, it's an error."""

        self.interactive = interactive
        self.dirname = ""
        self.filename = ""
        self.fullpath = ""

        if os.path.isfile( initpath ):
            # it's a file
            self.fullpath = initpath
            self.dirname, self.filename = os.path.split( self.fullpath )

        elif self.interactive:
            # it's a directory -- ask for a filename in that directory

            # make a list of available file extensions, with selected default first
            extensions = {'.fmf': 'fly movie format files (*.fmf)',
                          '.avi': 'audio-video interleave files (*.avi)',
                          '.sbfmf': 'static background fly movie format Files (*.sbfmf)',
                          '.ufmf': 'micro fly movie format Files (*.ufmf)'}
            if len( known_extensions() ) != len( extensions ):
                print "movie-open dialog doesn't list the same number of extensions as known_extensions()"
                
            dialog_str = ''
            # dlg.SetFilterIndex() could do this, too
            if default_extension in extensions.keys():
                dialog_str = extensions[default_extension] + '|*' + default_extension + '|'
                del extensions[default_extension]
            for ext, txt in extensions.iteritems():
                dialog_str += txt + '|*' + ext + '|'
            dialog_str += 'Any (*)|*'

            # show dialog and get selected filename
            dlg = wx.FileDialog( parentframe, "Select movie", initpath, "", dialog_str, wx.OPEN )
        
            if dlg.ShowModal() == wx.ID_OK:
                self.fullpath = dlg.GetPath()
                self.dirname, self.filename = os.path.split( self.fullpath )
            else:
                raise ImportError( "no filename was selected" )
                
            dlg.Destroy()

        else:
            raise ValueError( "not in interactive mode but wasn't given a full filename" )

        if open_now:
            self.open()


    def open( self ):
        """Figure out file type and initialize reader."""

        (front, ext) = os.path.splitext( self.fullpath )
        ext = ext.lower()
        if ext not in known_extensions():
            print "unknown file extension; will try AVbin to open"

        # read FlyMovieFormat
        if ext == '.fmf':
            self.type = 'fmf'
            try:
                self.h_mov = fmf.FlyMovie( self.fullpath )
            except NameError:
                if self.interactive:
                    wx.MessageBox( "Couldn't open \"%s\"\n(maybe FMF is not installed?)"%(filename), "Error", wx.ICON_ERROR|wx.OK )
                raise
            except IOError:
                if self.interactive:
                    wx.MessageBox( "I/O error opening \"%s\""%(self.fullpath), "Error", wx.ICON_ERROR|wx.OK )
                raise
        # read Static Background FlyMovieFormat
        elif ext == '.sbfmf':
            self.type = 'sbfmf'
            try:
                self.h_mov = fmf.FlyMovie( self.fullpath )
            except NameError:
                if self.interactive:
                    wx.MessageBox( "Couldn't open \"%s\"\n(maybe FMF is not installed?)"%(filename), "Error", wx.ICON_ERROR|wx.OK )
                raise
            except IOError:
                if self.interactive:
                    wx.MessageBox( "I/O error opening \"%s\""%(self.fullpath), "Error", wx.ICON_ERROR|wx.OK )
                raise
        # read Micro FlyMovieFormat
        elif ext == '.ufmf':
            self.type = 'ufmf'
            try:
                self.h_mov = ufmf.FlyMovieEmulator( self.fullpath )
            except NameError:
                if self.interactive:
                    wx.MessageBox( "Couldn't open \"%s\"\n(maybe UFMF is not installed?)"%(filename), "Error", wx.ICON_ERROR|wx.OK )
                raise
            except IOError:
                if self.interactive:
                    wx.MessageBox( "I/O error opening \"%s\""%(self.fullpath), "Error", wx.ICON_ERROR|wx.OK )
                raise
            except ufmf.ShortUFMFFileError:
                if self.interactive:
                    wx.MessageBox( "Error opening \"%s\". Short ufmf file."%(filename), "Error", wx.ICON_ERROR|wx.OK )
                raise
            except ufmf.CorruptIndexError:
                if self.interactive:
                    wx.MessageBox( "Error opening \"%s\". Corrupt file index."%(filename), "Error", wx.ICON_ERROR|wx.OK )
                raise
            except ufmf.InvalidMovieFileException:
                if self.interactive:
                    wx.MessageBox( "Error opening \"%s\". Invalid movie file."%(filename), "Error", wx.ICON_ERROR|wx.OK )
                raise
        # read AVI
        elif ext == '.avi':
            try:
                # KB: if use_uncompressed_avi set to False, then
                # only try CompressedAvi class
                if not params.use_uncompressed_avi:
                    if DEBUG: print "Not using uncompressed AVI class"
                    raise
                self.h_mov = Avi( self.fullpath )
                self.type = 'avi'
            except:
                try:
                    self.h_mov = CompressedAvi( self.fullpath )
                    self.type = 'avbin'
                except Exception, details:
                    if self.interactive:
                        if USE_AVBIN and not media.have_avbin:
                            msgtxt = "Failed opening file \"%s\". AVbin could not be loaded, and compressed AVIs cannot be read."%( self.fullpath )
                        else:
                            msgtxt = "Failed opening file \"%s\". AVbin was successfully loaded, but could not read the AVI."%( self.fullpath )
                        wx.MessageBox( msgtxt, "Error", wx.ICON_ERROR|wx.OK )
                    raise
                else:
                    if self.interactive:
                        wx.MessageBox("Your movie is most likely compressed. Out-of-order frame access (e.g. dragging the frame slider toolbars around) will be slow. At this time, the frame chosen to be displayed may be off by one or two frames, i.e. may not line up perfectly with computed trajectories.","Warning",wx.ICON_WARNING|wx.OK)
                    else:
                        print "reading compressed AVI"

            if self.interactive and self.h_mov.bits_per_pixel == 24 and not DEBUG_MOVIES and False:
                wx.MessageBox( "Currently, RGB movies are immediately converted to grayscale. All color information is ignored.", "Warning", wx.ICON_WARNING|wx.OK )

        # unknown movie type
        else:
            try:
                self.h_mov = CompressedAvi( self.fullpath )
                self.type = 'avbin'
            except:
                if self.interactive:
                    wx.MessageBox( "Failed opening file \"%s\"."%(self.fullpath), "Error", wx.ICON_ERROR|wx.OK )
                raise
            else:
                if self.interactive:
                    wx.MessageBox("Ctrax is assuming your movie is in an AVI format and is most likely compressed. Out-of-order frame access (e.g. dragging the frame slider toolbars around) will be slow. At this time, the frame chosen to be displayed may be off by one or two frames, i.e. may not line up perfectly with computed trajectories.","Warning",wx.ICON_WARNING|wx.OK)
                else:
                    print "reading movie as compressed AVI"

        self.file_lock = multiprocessing.RLock()
                    
        # add a buffer of the current frame
        self.bufferedframe_im = None
        self.bufferedframe_stamp = None
        self.bufferedframe_num = None


    def is_open( self ):
        return hasattr( self, 'h_mov' )


    def close( self ):
        """Close the movie file."""
        del self.file_lock
        del self.h_mov
        del self.type
        

    def get_frame( self, framenumber ):
        """Return numpy array containing frame data."""
        # check to see if we have buffered this frame
        if framenumber == self.bufferedframe_num:
            return (self.bufferedframe_im.copy(),self.bufferedframe_stamp)

        with self.file_lock:
            try:
                frame, stamp = self.h_mov.get_frame( framenumber )
            except (IndexError, NoMoreFramesException):
                if self.interactive:
                    wx.MessageBox( "Frame number %d out of range"%(framenumber), "Error", wx.ICON_ERROR|wx.OK )
                else:
                    print "frame", framenumber, "out of range"
                raise
            except (ValueError, AssertionError):
                if self.interactive:
                    wx.MessageBox( "Error reading frame %d"%(framenumber), "Error", wx.ICON_ERROR|wx.OK )
                else:
                    print "error reading frame", framenumber
                raise
            else:
                # store the current frame in the buffer
                self.bufferedframe_im = frame.copy()
                self.bufferedframe_stamp = stamp
                self.bufferedframe_num = framenumber

                return frame, stamp


    def get_n_frames( self ):
        with self.file_lock:
            return self.h_mov.get_n_frames()
    def get_width( self ):
        with self.file_lock:
            return self.h_mov.get_width()
    def get_height( self ):
        with self.file_lock:
            return self.h_mov.get_height()

    def get_some_timestamps( self, t1=0, t2=num.inf ):
        with self.file_lock:
            t2 = min(t2,self.get_n_frames())
            timestamps = self.h_mov.get_all_timestamps()
            timestamps = timestamps[t1:t2]
            return timestamps


    def writesbfmf_start(self,bg,filename):

        # write from start_frame to nframes-1
        self.nframescompress = self.get_n_frames() - params.start_frame

        # allocate array to store the addresses of each frame
        self.writesbfmf_framestarts = num.zeros(self.nframescompress)

        # open the output file
        self.writesbfmf_outfilename = filename
        self.outfile = open(self.writesbfmf_outfilename,'wb')

        # write the header
        self.writesbfmf_writeheader(bg)

    def writesbfmf_isopen(self):
        if not hasattr( self, 'outfile' ) or self.outfile is None:
            return False
        return (not self.outfile.closed)

    def writesbfmf_restart(self,frame,bg,filename):

        self.outfile = None

        self.writesbfmf_outfilename = filename

        self.nframescompress = self.get_n_frames() - params.start_frame

        # allocate array to store the addresses of each frame
        self.writesbfmf_framestarts = num.zeros(self.nframescompress)

        # move the file to a temporary file
        tmpfilename = 'tmp_ctrax_writesbfmf.sbfmf'
        os.rename(filename,tmpfilename)

        # open the old file for reading
        inmovie = Movie( tmpfilename, self.interactive )
            
        # open the output file
        self.outfile = open(filename,"wb")
        
        # rewrite the header
        self.writesbfmf_writeheader(bg)

        # last frame to copy
        i = frame - params.start_frame - 1
        firstaddr = inmovie.h_mov.framelocs[0]
        lastaddr = inmovie.h_mov.framelocs[i]

        self.writesbfmf_framestarts[:i+1] = inmovie.h_mov.framelocs[:i+1]

        if DEBUG_MOVIES: print "copied framestarts: "
        if DEBUG_MOVIES: print str(self.writesbfmf_framestarts[:i+1])

        # seek to the first frame
        inmovie.h_mov.seek(0)

        # copy in pages of size pagesize
        pagesize = int(2**20)
        for j in range(firstaddr,lastaddr,pagesize):
            if DEBUG_MOVIES: print "writing page at %d"%inmovie.h_mov.file.tell()
            buf = inmovie.h_mov.read_some_bytes(pagesize)
            self.outfile.write(buf)

        # write last page
        if j < lastaddr:
            if DEBUG_MOVIES: print "writing page at %d"%inmovie.h_mov.file.tell()
            buf = inmovie.h_mov.read_some_bytes(int(lastaddr-pagesize))
            self.outfile.write(buf)

        # close the input movie and delete
        inmovie.h_mov.close()
        os.remove(tmpfilename)


    def writesbfmf_close(self,frame):
        if hasattr( self, 'outfile' ) and self.outfile is not None:
            # write the index
            self.writesbfmf_writeindex(frame)

            # close the file
            self.outfile.close()


    def writesbfmf_writeindex(self,frame):
        """
        Writes the index at the end of the file. Index consists of nframes unsigned long longs (Q),
        indicating the positions of each frame
        """
        # write the index
        indexloc = self.outfile.tell()
        nframeswrite = frame - params.start_frame + 1
        if DEBUG_MOVIES: print "writing index, nframeswrite = %d"%nframeswrite
        for i in range(nframeswrite):
            self.outfile.write(struct.pack("<Q",self.writesbfmf_framestarts[i]))

        # write the location of the index
        self.outfile.seek(self.writesbfmf_indexptrloc)
        self.outfile.write(struct.pack("<Q",indexloc))            

        # write the number of frames
        self.outfile.seek(self.writesbfmf_nframesloc)
        self.outfile.write(struct.pack("<I",nframeswrite))


    def writesbfmf_writeheader(self,bg):
        """
        Writes the header for the file. Format:
        Number of bytes in version string: (I = unsigned int)
        Version Number (string of specified length)
        Number of rows (I = unsigned int)
        Number of columns (I = unsigned int)
        Number of frames (I = unsigned int)
        Difference mode (I = unsigned int):
          0 if light flies on dark background, unsigned mode
          1 if dark flies on light background, unsigned mode
          2 if other, signed mode
        Location of index (Q = unsigned long long)
        Background image (ncols * nrows * double)
        Standard deviation image (ncols * nrows * double)
        """

        self.nr = self.get_height()
        self.nc = self.get_width()

        # write the number of columns, rows, frames, difference mode
        if bg.bg_type == 'light_on_dark':
            difference_mode = 0
        elif bg.bg_type == 'dark_on_light':
            difference_mode = 1
        else:
            difference_mode = 2
        self.outfile.write(struct.pack("<I",len(__version__)))
        self.outfile.write(__version__)
        self.outfile.write(struct.pack("<2I",int(self.nr),int(self.nc)))
        self.writesbfmf_nframesloc = self.outfile.tell()
        self.outfile.write(struct.pack("<2I",int(self.nframescompress),int(difference_mode)))

        if DEBUG_MOVIES: print "writeheader: nframescompress = " + str(self.nframescompress)

        # compute the location of the standard deviation image
        stdloc = self.outfile.tell() + struct.calcsize("B")*self.nr*self.nc

        # compute the location of the first frame
        ffloc = stdloc + struct.calcsize("d")*self.nr*self.nc

        # where do we write the location of the index -- this is always the same
        self.writesbfmf_indexptrloc = self.outfile.tell()

        # write a placeholder for the index location
        self.outfile.write(struct.pack("<Q",0))

        # write the background image
        self.outfile.write(bg.center)

        # write the standard deviation image
        self.outfile.write(bg.dev)
        

    def writesbfmf_writeframe(self,isfore,im,stamp,currframe):

        if DEBUG_MOVIES: print "writing frame %d"%currframe

        tmp = isfore.copy()
        tmp.shape = (self.nr*self.nc,)
        i, = num.nonzero(tmp)

        # values at foreground pixels
        v = im[isfore]

        # number of foreground pixels
        n = len(i)

        # store the start of this frame
        j = currframe - params.start_frame
        self.writesbfmf_framestarts[j] = self.outfile.tell()

        if DEBUG_MOVIES: print "stored in framestarts[%d]"%j

        # write number of pixels and time stamp
        self.outfile.write(struct.pack("<Id",n,stamp))

        i = i.astype(num.uint32)

        self.outfile.write(i)
        self.outfile.write(v)
        

    def close(self):
        if hasattr(self,'h_mov'):
            with self.file_lock:
                try:
                    self.h_mov.close()
                except:
                    print "Could not close"


"""
AVI class; written by JAB and KMB, altered by Don Olbris.

Don's changes:
important alterations from version I received:
- allows fccHandler = "" or all nulls
- allows width to vary to match actual frame size (suspect that avi pads to 
    multiples of four?) (and let get_width return actual width)
- added various derived attributes to better masquerade as an FMF file
- added an "fmf" mode; in this mode, it reshapes array to same shape as fmf
    (row/column order swapped) -> was commented out, so removed JAB 8/29/11
- added a seek() method

- I want to make this width change more transparent, but I keep
    running into related shape issues

- Avi class is still byte-order undefined, but at least it's read-only
"""

class Avi:
    """Read uncompressed AVI movies."""
    def __init__( self, filename ):

        self.issbfmf = False
 
        # need to open in binary mode to support Windows:
        self.file = open( filename, 'rb' )

        self.frame_index = {} # file locations of each frame

        try:
            self.read_header()
            self.postheader_calculations()
        except Exception, details:
            if DEBUG_MOVIES: print( "error reading uncompressed AVI:" )
            if DEBUG_MOVIES: print( details )
            raise
        
        # added to help masquerade as FMF file:
        self.filename = filename
        self.chunk_start = self.data_start
        # this is a mystery, but I think it's 8 for avi: seems to be offset
        #   from beginning of "chunk" to beginning of array data within chunk
        self.timestamp_len = 8
        if hasattr(self, "newwidth"):
            self.bytes_per_chunk = (self.height * self.newwidth) + self.timestamp_len
        else:
            self.bytes_per_chunk = self.buf_size + self.timestamp_len
        #self.bits_per_pixel = 8
        if DEBUG_MOVIES: print "bits per pix: %d, bytes per chunk %d" % (self.bits_per_pixel, self.bytes_per_chunk)


    def get_all_timestamps( self ):
        """Return a Numpy array containing all frames' timestamps."""
        timestamps = num.zeros( (self.n_frames,) )
        for fr in range( self.n_frames ):
            timestamps[fr] = self.make_timestamp( fr )
        return timestamps
    
        
    def make_timestamp( self, fr ):
        """Approximate timestamp from frame rate recorded in header."""
        if self.frame_delay_us != 0:
            return fr * self.frame_delay_us / 1e6
        elif self.time_scale != 0:
            return fr * self.data_rate / float(self.time_scale)
        else:
            return fr / 30.


    ###################################################################
    # read_header()
    ###################################################################
    def read_header( self ):

        # read RIFF then riffsize
        RIFF, riff_size, AVI = struct.unpack( '4sI4s', self.file.read( 12 ) )
        if not RIFF == 'RIFF':
            print "movie header RIFF error at", RIFF, riff_size, AVI
            raise TypeError("Invalid AVI file. Must be a RIFF file.")
        if (not AVI == 'AVI ') and (not AVI == 'AVIX'):
            print "movie header AVI error at", RIFF, riff_size, AVI
            raise TypeError("Invalid AVI file. File type must be \'AVI \'.")

        # read hdrl
        LIST, hdrl_size, hdrl = struct.unpack( '4sI4s', self.file.read( 12 ) )
        hdrlstart = self.file.tell() - 4

        if not LIST == 'LIST':
            print "movie header LIST 1 error at", LIST, hdrl_size, hdrl
            raise TypeError("Invalid AVI file. Did not find header list.")

        if hdrl == 'hdrl': # a real header
            # read avih 
            avih, avih_size = struct.unpack( '4sI', self.file.read( 8 ) )
            if not avih == 'avih':
                print "movie header avih error at", avih, avih_size
                raise TypeError("Invalid AVI file. Did not find avi header.")
            avihchunkstart = self.file.tell()

            # read microsecperframe
            self.frame_delay_us, = struct.unpack('I',self.file.read(4))

            # skip to nframes
            self.file.seek(3*4,1)
            self.n_frames, = struct.unpack('I',self.file.read(4))
            
            # skip to width, height
            self.file.seek(3*4,1)
            self.width,self.height = struct.unpack('2I',self.file.read(8))

            if DEBUG_MOVIES: print "width = %d, height = %d"%(self.width,self.height)
            if DEBUG_MOVIES: print "n_frames = %d"%self.n_frames

            # skip the rest of the aviheader
            self.file.seek(avihchunkstart+avih_size,0)

            LIST, stream_listsize, strl = \
                  struct.unpack( '4sI4s', self.file.read( 12 ) )

            if (not LIST == 'LIST') or (not strl == 'strl'):
                print "movie header LIST 2 error at", LIST, strl
                raise TypeError("Invalid AVI file. Did not find stream list.")

            strh, strh_size = struct.unpack( '4sI', self.file.read( 8 ) )
            if not strh == 'strh':
                print "movie header strh error at", strh, strh_size
                raise TypeError("Invalid AVI file. Did not find stream header.")

            strhstart = self.file.tell()

            # read stream type, fcc handler
            vids, fcc = struct.unpack( '4s4s', self.file.read( 8 ) )
            # check for vidstream
            if not vids == 'vids':
                print "movie header vids error at", vids
                raise TypeError("Unsupported AVI file type. First stream found is not a video stream.")
            # check fcc
            if fcc not in ['DIB ', '\x00\x00\x00\x00', "", "RAW ", "NONE", chr(24)+"BGR", 'Y8  ']:
                if DEBUG_MOVIES: print "movie header codec error at", fcc
                raise TypeError("Unsupported AVI file type %s, only uncompressed AVIs supported."%fcc)
            if DEBUG_MOVIES: print "codec", fcc

            # skip the rest of the stream header
            self.file.seek(strhstart+strh_size,0)

            strf, strf_size = struct.unpack( '4sI', self.file.read( 8 ) )
            if not strf == "strf":
                print "movie header strf error at", strf
                raise TypeError("Invalid AVI file. Did not find strf.")

            strfstart = self.file.tell()
            bitmapheadersize, = struct.unpack('I',self.file.read(4))

            # skip width, height, planes
            self.file.seek(4*2+2,1)

            # read in bits per pixel
            self.bits_per_pixel, = struct.unpack('H',self.file.read(2))
            if DEBUG_MOVIES: print "bits_per_pixel = %d"%self.bits_per_pixel

            # is this an indexed avi?
            colormapsize = (strf_size - bitmapheadersize)/4
            if colormapsize > 0:
                self.isindexed = True
                self.file.seek(strfstart+bitmapheadersize,0)
                self.colormap = num.frombuffer(self.file.read(4*colormapsize),num.uint8)
                self.colormap = self.colormap.reshape((colormapsize,4))
                self.colormap = self.colormap[:,:-1]
                if DEBUG_MOVIES: print "file is indexed with %d colors" % len( self.colormap )
            else:
                self.isindexed = False

            if self.bits_per_pixel == 24:
                self.isindexed = False

            # skip the rest of the strf
            self.file.seek(hdrlstart+hdrl_size,0)

        else:
            # maybe this is a "second header" following an index... no hdrl
            self.file.seek( -12, os.SEEK_CUR )

        while True:
            # find LIST chunk
            LIST,movilist_size = struct.unpack( '4sI', self.file.read( 8 ) )
            if LIST == 'LIST':
                # find movi
                movi, = struct.unpack('4s',self.file.read(4))
                if DEBUG_MOVIES: print 'looking for movi, found ' + movi
                if movi == 'movi':
                    break
                else:
                    self.file.seek(-4,1)
            # found some other chunk, seek past
            self.file.seek(movilist_size,1)            

        if not movi == 'movi':
            raise TypeError("Invalid AVI file. Did not find movi, found %s."%movi)

        # read extra stuff
        while True:
            fourcc,chunksize, = struct.unpack('4sI',self.file.read(8))
            if DEBUG_MOVIES: print 'read fourcc=%s, chunksize=%d'%(fourcc,chunksize)
            if fourcc == '00db' or fourcc == '00dc':
                self.file.seek(-8,1)
                break
            self.file.seek(chunksize,1)
            
        self.buf_size = chunksize
        if DEBUG_MOVIES: print "chunk size: ", self.buf_size

        # check whether n_frames makes sense (sometimes headers lie)
        approx_file_len = self.buf_size*self.n_frames
        cur_pos = self.file.tell()
        self.file.seek( 0, os.SEEK_END )
        real_file_len = self.file.tell()
        self.file.seek( cur_pos )

        if real_file_len > approx_file_len*1.1:
            print "approximate file length %ld bytes, real length %ld"%(approx_file_len, real_file_len)
            old_n_frames = self.n_frames
            self.n_frames = int( num.floor( (real_file_len - cur_pos)/self.buf_size ) )
            print "guessing %d frames in movie, although header said %d"%(self.n_frames,old_n_frames)


    ###################################################################
    # postheader_calculations()
    ###################################################################
    def postheader_calculations( self ):
        """Tidy up some initialization, immediately after reading header."""
        
        self.data_start = self.file.tell()

        # figure out padding
        depth = self.bits_per_pixel/8
        unpaddedframesize = self.width*self.height*depth

        if unpaddedframesize == self.buf_size:
            # no padding
            self.padwidth = 0
            self.padheight = 0
        elif unpaddedframesize + self.width*depth == self.buf_size:
            self.padwidth = 0
            self.padheight = 1
        elif unpaddedframesize + self.height*depth == self.buf_size:
            self.padwidth = 1
            self.padheight = 0
        else:
            raise TypeError("Invalid AVI file. Frame size (%d) does not match width * height * bytesperpixel (%d*%d*%d)."%(self.buf_size,self.width,self.height,depth))

        if self.bits_per_pixel == 24:
            self.format = 'RGB'
        elif self.isindexed:
            self.format = 'INDEXED'
        elif self.bits_per_pixel == 8:
            self.format = 'MONO8'
        else:
            raise TypeError("Unsupported AVI type. bitsperpixel must be 8 or 24, not %d."%self.bits_per_pixel)

        if DEBUG_MOVIES: print "format = " + str(self.format)

        if self.n_frames == 0:
            loccurr = self.file.tell()
            self.file.seek(0,2)
            locend = self.file.tell()
            self.n_frames = int( num.floor( (locend - self.data_start) / (self.buf_size+8) ) )
            print "n frames = 0; setting to %d"%self.n_frames

        #Barbara Casillas & Victor Mireles, 5/12/13
    	#We try to read the last frame. If it is not accessible
    	# ("frame x out of range" error) then get_frame will fix self.n_frames.
    	#So, we just keep reading the last frame until there is no correction.
    	currentNframes = self.n_frames #(estimated num. of frames)
    	while True:
            frameNumber = self.n_frames - 1
            try:
                self.get_frame(frameNumber)
            except:
                pass
            if self.n_frames == currentNframes:
                break
            currentNframes = self.n_frames


    def nearest_indexed_frame( self, framenumber ):
        """Return nearest known frame index less than framenumber."""
        keys = self.frame_index.keys()
        keys.sort()
        nearest = None
        for key in keys:
            if framenumber > key:
                nearest = key
            else:
                break
        return nearest


    def build_index( self, to_fr ):
        """Build index successively up to a selected frame."""

        # find frame to start from
        near_idx = self.nearest_indexed_frame( to_fr )
        if near_idx is None:
            # no frames read, read first one
            self.file.seek( self.data_start, os.SEEK_SET )
            self.framenumber = 0
            try:
                frame, stamp = self.get_next_frame()
            except:
                if params.interactive:
                    pb.Destroy()
                raise
            else:
                from_fr = 0
        else:
            # start immediately after last known frame
            from_fr = near_idx

        # open progress bar
        if params.interactive and to_fr - from_fr > 10:
            show_pb = True
        else:
            show_pb = False
        if show_pb:
            pb = wx.ProgressDialog( 'Building Frame Index',
                                    "Calculating file location for frame %d"%to_fr,
                                    to_fr - from_fr,
                                    None,
                                    wx.PD_APP_MODAL|wx.PD_AUTO_HIDE|wx.PD_CAN_ABORT|wx.PD_REMAINING_TIME )

        # read frames until target frame is reached
        max_increment = 100
        increment = max_increment
        last_fr = from_fr
        this_fr = min( from_fr + increment, to_fr )
        failed_fr = last_fr
        while True:
            # try reading several frames into the future
            self.file.seek( (self.buf_size + 8)*(this_fr - last_fr) + self.frame_index[last_fr], os.SEEK_SET )
            self.framenumber = this_fr
            try:
                frame, stamp = self.get_next_frame()
                if this_fr >= to_fr:
                    break
            except:
                # reading failed -- try reading fewer frames into the future
                if increment == 1:
                    if show_pb:
                        pb.Destroy()
                    raise
                else:
                    increment = max( int( increment/10 ), 1 )
                    failed_fr = last_fr
                    this_fr = min( last_fr + increment, to_fr )
            else:
                # reading successful -- increment and repeat
                if increment != max_increment and this_fr - failed_fr >= 10*increment:
                    # if it's been a long time since failure, speed up again
                    increment = min( 10*increment, max_increment )
                last_fr = this_fr
                this_fr = min( last_fr + increment, to_fr )

            if show_pb:
                (keepgoing, skip) = pb.Update( this_fr - from_fr )
                if not keepgoing:
                    pb.Destroy()
                    raise IndexError( "didn't finish building index to %d"%to_fr )

        if show_pb:
            pb.Destroy()
            
        
    ###################################################################
    # get_frame()
    ###################################################################
    def get_frame( self, framenumber ):
        """Read frame from file and return as NumPy array."""

        if DEBUG_MOVIES: print "uncompressed get_frame(%d)"%framenumber

        if framenumber < 0: raise IndexError
        if framenumber >= self.n_frames: raise NoMoreFramesException
        
        self.framenumber = framenumber

        # read frame from file
        if framenumber in self.frame_index:
            if DEBUG_MOVIES: print "calling frame %d from index at %d"%(framenumber,self.frame_index[framenumber])
            self.file.seek( self.frame_index[framenumber], os.SEEK_SET )
            return self.get_next_frame()
        
        else:
            near_idx = self.nearest_indexed_frame( framenumber )
            if near_idx is not None:
                # offset from nearest indexed frame
                offset = framenumber - near_idx
                self.file.seek( (self.buf_size + 8)*offset + self.frame_index[near_idx], os.SEEK_SET )
            else:
                # offset from beginning of file
                self.file.seek( self.data_start + (self.buf_size+8)*framenumber )

            try:
                return self.get_next_frame()
            except ValueError:
                if framenumber == 0: raise

                self.build_index( framenumber )

                self.file.seek( self.frame_index[framenumber], os.SEEK_SET )
                return self.get_next_frame()


    ###################################################################
    # get_next_frame()
    ###################################################################
    def get_next_frame(self):
        """returns next frame"""

        # test for end of file
        file_data = self.file.read( 8 )
        if len( file_data ) != 8:
            cur_pos = self.file.tell()
            self.file.seek( 0, os.SEEK_END )
            if self.file.tell() >= cur_pos:
                raise IndexError( "error seeking frame %d -- file not readable"%self.framenumber )
            else:
                self.n_frames = self.framenumber - 1
                self.framenumber = self.n_frames
                return self.get_next_frame()
        
        this_frame_id, frame_size = struct.unpack( '4sI', file_data )
        if DEBUG_MOVIES: print 'frame id=%s, sz=%d'%(this_frame_id,frame_size)

        if this_frame_id == 'idx1' or \
               this_frame_id == 'ix00' or \
               this_frame_id == 'ix01': # another index midstream
            a = self.file.read( frame_size )
            this_frame_id, frame_size = struct.unpack( '4sI', self.file.read( 8 ) )
            if DEBUG_MOVIES: print 'skipped index; frame id=' + str(this_frame_id) + ', sz=' + str(frame_size)

        if this_frame_id == 'RIFF': # another whole header
            self.file.seek( -8, os.SEEK_CUR )
            self.read_header()
            this_frame_id, frame_size = struct.unpack( '4sI', self.file.read( 8 ) )
            if DEBUG_MOVIES: print 'skipped another header; frame id=' + str(this_frame_id) + ', sz=' + str(frame_size)

        if hasattr( self, 'frame_id' ) and this_frame_id != self.frame_id:
            # who knows? try skipping ahead a bit
            tries = 0
            while this_frame_id != self.frame_id and tries < 64:
                self.file.seek( -7, os.SEEK_CUR )
                this_frame_id, frame_size = struct.unpack( '4sI', self.file.read( 8 ) )
                tries += 1
            if DEBUG_MOVIES: print "skipped forward %d bytes; now id=%s, sz=%d"%(tries,this_frame_id, frame_size)

        if frame_size != self.buf_size:
            raise ValueError( "Frame size %d on disk does not equal uncompressed size %d; movie must be uncompressed"%(frame_size, self.buf_size) )
        if not hasattr( self, 'frame_id' ):
            self.frame_id = this_frame_id
        elif this_frame_id != self.frame_id:
            if DEBUG_MOVIES: print "looking for header %s; found %s"%(self.frame_id,this_frame_id)
            raise ValueError( "error seeking frame start: unknown data header" )

        # make frame into numpy array
        frame_data = self.file.read( frame_size )
        frame = num.fromstring( frame_data, num.uint8 )

        # reshape...
        width = self.width + self.padwidth
        height = self.height + self.padheight
        if self.isindexed:
            frame = self.colormap[frame,:]
            if params.movie_index_transpose:
                # JAB 20130304: old code always executed this, but it's
                # wrong for the one indexed movie type I have available
                frame.resize((width,height,3))
                frame = frame[:self.width,:self.height,:]
            else:
                frame.resize( (height, width, 3) )
                frame = frame[:self.height,:self.width,:]
            tmp = frame.astype(float)
            tmp = tmp[:,:,0]*.3 + tmp[:,:,1]*.59 + tmp[:,:,2]*.11 # RGB -> L
            if params.movie_index_transpose:
                tmp = tmp.T
            frame = tmp.astype(num.uint8)
        elif frame.size == width*height:
            frame.resize( (height, width) )
            frame = frame[:self.height,:self.width]
        elif frame.size == width*height*3:
            frame.resize( (height, width*3) )
            tmp = frame.astype(float)
            tmp = tmp[:,2:width*3:3]*.3 + \
                tmp[:,1:width*3:3]*.59 + \
                tmp[:,0:width*3:3]*.11 # RGB -> L
            tmp = tmp[:self.height,:self.width]
            frame = tmp.astype(num.uint8)
            #frame = imops.to_mono8( 'RGB24', frame )
            #raise TypeError( "movie must be grayscale" )
        else:
            # frame size doesn't match; for this exercise, pretend the height is 
            #   right and see if width is integral and within 10 of expected;
            #   if so, use that; otherwise, error (djo)
            # raise ValueError( "frame size %d doesn't make sense: movie must be 8-bit grayscale"%(frame.size) )
            if frame.size % height == 0:
                self.newwidth = frame.size / height
                if abs(self.newwidth - width) < 10:
                    frame.resize((self.newwidth, height))
                    frame = frame[:self.width,:self.height]
                else:
                    raise ValueError("apparent new width = %d; expected width = %d"
                        % (height, self.newwidth))
            else:
                print self.width, self.height, self.padwidth, self.padheight
                print self.width*self.height, frame_size, frame.size, self.width*self.height*3, frame_size/3
                print frame_size/self.width/3, frame_size/self.height/3, frame_size % width, frame_size % height
                raise ValueError("apparent new width is not integral; mod = %d" % (frame.size % height))
            
        if self.framenumber not in self.frame_index:
            self.frame_index[self.framenumber] = self.file.tell() - frame_size - 8
            if DEBUG_MOVIES: print "added frame %d to index at %d"%(self.framenumber,self.frame_index[self.framenumber])
        
        return frame, self.make_timestamp( self.framenumber )
        
        # end get_next_frame()
    
    def get_n_frames( self ): 
        return self.n_frames
    
    def get_width( self ): 
        if hasattr(self, "newwidth"):
            return self.newwidth
        else:
            return self.width
    
    def get_height( self ): 
        return self.height
    
    def seek(self,framenumber):
        if framenumber < 0:
            framenumber = self.n_frames + framenumber
        if framenumber in self.frame_index:
            seek_to = self.frame_index[framenumber]
        else:
            seek_to = self.chunk_start + self.bytes_per_chunk*framenumber
        self.file.seek(seek_to)
    
    # end class Avi


class CompressedAvi:

    """Use pyglet.media to read compressed avi files"""
    def __init__(self,filename):

        if not USE_AVBIN:
            raise Exception, 'Trying to read compressed AVI, but USE_AVBIN flag set to False'

        if DEBUG_MOVIES: print 'Trying to read compressed AVI'
        self.issbfmf = False
        self.source = media.load(filename)
        if not hasattr( self.source, '_seek' ):
            # Pyglet >= 1.2
            self.source._seek = self.source.seek
        self.ZERO = 1./1000000.

        # for some video types, if we don't do a read before 
        # seeking, bad things happen
        im = self.source.get_next_video_frame()
        ts = self.source.get_next_video_timestamp()
        if DEBUG_MOVIES: print 'first frame read results in im = ' + str(im)
        if DEBUG_MOVIES: print 'and ts = ' + str(ts)
        self.source._seek(self.ZERO)
        self.start_time = self.source.get_next_video_timestamp()
        if DEBUG_MOVIES: print 'First seek to ZERO succeeded, start_time = ' + str(self.start_time)

        # estimate properties of the video that would be nice 
        # to be able to read in -- i have no idea how accurate 
        # these estimates are
        self.duration_seconds = self.source.duration
        if DEBUG_MOVIES: print 'duration_seconds = ' + str(self.duration_seconds)
        self._estimate_fps()
        if DEBUG_MOVIES: print 'fps estimated to be' + str(self.fps)
        self.n_frames = int(num.floor(self.duration_seconds * self.fps))
        if DEBUG_MOVIES: print "n_frames estimated to be %d"%self.n_frames
        self.frame_delay_us = 1e6 / self.fps
        self._estimate_keyframe_period()
        if DEBUG_MOVIES: print 'keyframe period estimate to be ' + str(self.keyframe_period_s) + ' s, ' + str(self.keyframe_period) + ' frames'
        
        # added to help masquerade as FMF file:
        self.filename = filename

        # read in the width and height of each frame
        self.width = self.source.video_format.width
        self.height = self.source.video_format.height

        self.MAXBUFFERSIZE = num.round(200*1000*1000/self.width/self.height)
        self.buffersize = min(self.MAXBUFFERSIZE,self.keyframe_period)
        if DEBUG_MOVIES: print 'buffersize set to ' + str(self.buffersize)

        # compute the bits per pixel
        self.source._seek(self.ZERO)
        im = self.source.get_next_video_frame()
        im = num.fromstring(im.data,num.uint8)
        self.color_depth = len(im)/self.width/self.height
        if self.color_depth != 1 and self.color_depth != 3:
            raise ValueError( 'color_depth = %d, only know how to deal with color_depth = 1 or colr_depth = 3'%self.color_depth )
        self.bits_per_pixel = self.color_depth * 8

        # allocate the buffer
        self.buffer = num.zeros((self.height,self.width,self.buffersize),dtype=num.uint8)
        self.bufferts = num.zeros(self.buffersize)
        # put the first frame in it
        self.source._seek(self.ZERO)
        self.currframe = 0
        (frame,ts) = self.get_next_frame_and_reset_buffer()

        if DEBUG_MOVIES: print "Done initializing CompressedAVI"

    def get_all_timestamps( self ):
        return num.arange( self.n_frames )/self.fps + self.start_time
        

    def get_frame(self,framenumber):
        """Read frame from file and return as NumPy array."""

        if framenumber < 0: raise IndexError

        # have we already stored this frame?
        if framenumber >= self.bufferframe0 and framenumber < self.bufferframe1:
            off = num.mod(framenumber - self.bufferframe0 + self.bufferoff0,self.buffersize)
            if DEBUG_MOVIES: print "frame %d is in buffer at offset %d"%(framenumber,off)
            return (self.buffer[:,:,off].copy(),self.bufferts[off])

        # is framenumber the next frame to read in?
        if framenumber == self.currframe:
            if DEBUG_MOVIES: print "frame %d is the next frame, just calling get_next_frame"%framenumber
            return self.get_next_frame()

        # otherwise, we need to seek

        # find the last keyframe at or before framenumber
        lastkeyframe = num.floor(framenumber / self.keyframe_period)*self.keyframe_period
        lastkeyframe_s = lastkeyframe / self.fps + self.start_time

        if DEBUG_MOVIES: print "keyframe before frame %d is %d (ts = %f)"%(framenumber,lastkeyframe,lastkeyframe_s)

        didbeginbusy = False
        if params.interactive:
            try:
                wx.Yield()
            except:
                pass
            else:
                if not wx.IsBusy():
                    didbeginbusy = True
                    wx.BeginBusyCursor()

        # is the current frame before the desired frame and at or after the nearest keyframe?
        if framenumber > self.currframe and self.currframe >= lastkeyframe:

            if DEBUG_MOVIES: print "the best thing to do is just to read in frames until we get to desired %d (currframe = %d)"%(framenumber,self.currframe)

            # read into the buffer
            for f in range(self.currframe,framenumber+1):
                try:
                    (frame,ts) = self.get_next_frame()
                except:
                    if didbeginbusy:
                        wx.EndBusyCursor()
                    raise
            
            if didbeginbusy:
                wx.EndBusyCursor()

            return (frame,ts)
            
        # seek around til we find the right window

        if DEBUG_MOVIES: print "we need to seek"

        # make sure we don't move in opposite directions ever
        dirmoved = 0

        # convert from frame to time
        wantts = framenumber / self.fps + self.start_time
        if DEBUG_MOVIES: print "frame %d is equivalent to time stamp %f"%(framenumber,wantts)
        while True:
            # try seeking
            if DEBUG_MOVIES: print "start_time = %f"%self.start_time
            if DEBUG_MOVIES: print "trying to seek to %f (actually %f)"%(lastkeyframe_s,lastkeyframe_s-self.start_time)
            try: 
                self.source._seek(lastkeyframe_s-self.start_time)
            except:
                print 'seeking failed, aborting. tried to seek to %f (actually %f)'%(lastkeyframe_s,lastkeyframe_s-self.start_time)
            tscurr = self.source.get_next_video_timestamp()
            if tscurr is None:
                print 'seeking failed, aborting. timestamp returned is None'
                raise NoMoreFramesException

            if DEBUG_MOVIES: print "ended up seeking to %f"%tscurr

            # are we in the right ballpark?
            if wantts < tscurr:

                lastkeyframe_s = max(self.ZERO,lastkeyframe_s-self.keyframe_period_s)
                if DEBUG_MOVIES: print "still not back far enough, trying to seek to %f"%lastkeyframe_s
                dirmoved = -1
                continue

            elif wantts >= tscurr + self.keyframe_period_s:
                if dirmoved == -1:
                    # we went back and now we think we need to go forward
                    # so we're screwed up
                    # let's just stay
                    if DEBUG_MOVIES: print "uh oh -- we've already gone backward, and now we think we need to go forward, so we're just going to stay"
                    break
                lastkeyframe_s += self.keyframe_period_s
                dirmoved = 1
                continue
            else:
                # we're in the right ballpark
                if DEBUG_MOVIES: print "we're in the right ballpark"
                break

        # end loop searching for keyframe

        lastkeyframe = (tscurr - self.start_time)*self.fps
        self.currframe = int(lastkeyframe)

        (frame,ts) = self.get_next_frame_and_reset_buffer()

        # go forward until we get the right frame
        for f in range(self.currframe,framenumber+1):
            try:
                (frame,ts) = self.get_next_frame()
            except:
                if didbeginbusy:
                    wx.EndBusyCursor()
                raise

        if didbeginbusy:
            wx.EndBusyCursor()

        # return
        return (frame,ts)

    def get_next_frame_and_reset_buffer(self):

        # first frame stored in buffer
        self.bufferframe0 = self.currframe
        # frame after last frame stored in buffer
        self.bufferframe1 = self.currframe + 1
        # buffer will wrap around if frames are read in in-order
        # bufferoff0 is the location in the buffer of the first 
        # frame buffered
        self.bufferoff0 = 0
        (frame,ts) = self._get_next_frame_helper()
        self.buffer[:,:,0] = frame.copy()
        self.bufferts[0] = ts
        # bufferoff is the location in the buffer where we should 
        # write the next frame
        if self.buffersize > 1:
            self.bufferoff = 1
        else:
            self.bufferoff = 0

        # set the current frame
        self.currframe += 1
        self.prevts = ts

        return (frame,ts)

    def _get_next_frame_helper(self):
        ts = self.source.get_next_video_timestamp()
        if ts is None:
            if self.currframe - 1 < self.n_frames:
                self.n_frames = self.currframe - 1
            raise NoMoreFramesException
        im = self.source.get_next_video_frame()
        frame = num.fromstring(im.data,num.uint8)

        if self.color_depth == 1:
            frame.resize((im.height,im.width))
        else: # color_depth == 3
            frame.resize( (self.height, self.width*3) )
            tmp = frame.astype(float)
            tmp = tmp[:,2:self.width*3:3]*.3 + \
                tmp[:,1:self.width*3:3]*.59 + \
                tmp[:,0:self.width*3:3]*.11
            frame = tmp.astype(num.uint8)

        frame = num.flipud(frame)

        return (frame,ts)

    def get_next_frame(self):

        (frame,ts) = self._get_next_frame_helper()

        # store 

        # last frame stored will be 1 more
        self.bufferframe1 += 1

        # are we erasing the first frame?
        if self.bufferoff0 == self.bufferoff:
            self.bufferframe0 += 1
            if self.buffersize > 1:
                self.bufferoff0 += 1
            if DEBUG_MOVIES: print "erasing first frame, bufferframe0 is now %d, bufferoff0 is now %d"%(self.bufferframe0,self.bufferoff0)

        if DEBUG_MOVIES: print "buffer frames: [%d,%d), bufferoffset0 = %d"%(self.bufferframe0,self.bufferframe1,self.bufferoff0)
            
        self.buffer[:,:,self.bufferoff] = frame.copy()
        self.bufferts[self.bufferoff] = ts

        if DEBUG_MOVIES: print "read into buffer[%d], ts = %f"%(self.bufferoff,ts)

        self.bufferoff += 1

        # wrap around
        if self.bufferoff >= self.buffersize:
            self.bufferoff = 0

        if DEBUG_MOVIES: print "incremented bufferoff to %d"%self.bufferoff

        # remember current location in the movie
        self.currframe += 1
        self.prevts = ts

        if DEBUG_MOVIES: print "updated currframe to %d, prevts to %f"%(self.currframe,self.prevts)

        return (frame,ts)

    def _estimate_fps(self):

        if DEBUG_MOVIES: print 'Estimating fps'

        # seek to the start of the stream
        self.source._seek(self.ZERO)

        if DEBUG_MOVIES: print 'First seek succeeded'

        # initial time stamp
        ts0 = self.source.get_next_video_timestamp()
        ts1 = ts0

        if DEBUG_MOVIES: print 'initial time stamp = ' + str(ts0)

        # get the next frame and time stamp a bunch of times
        nsamples = 200
        if DEBUG_MOVIES: print 'nsamples = ' + str(nsamples)
        i = 0 # i is the number of frames we have successfully grabbed
        while True:
            im = self.source.get_next_video_frame()
            ts = self.source.get_next_video_timestamp()
            if DEBUG_MOVIES: print 'i = %d, ts = '%i + str(ts)
            if (ts is None) or num.isnan(ts) or (ts <= ts1):
                break
            i = i + 1
            ts1 = ts
            if i >= nsamples:
                break
        
        if ts1 <= ts0:
            raise ValueError( "Could not compute the fps in the compressed movie" )

        self.fps = float(i) / (ts1-ts0)
        if DEBUG_MOVIES: print 'Estimated frames-per-second = %f'%self.fps
        
    def _estimate_keyframe_period(self):

        if DEBUG_MOVIES: print 'Estimating keyframe period'

        self.source._seek(self.ZERO)

        ts0 = self.source.get_next_video_timestamp()

        if DEBUG_MOVIES: print 'After first seek, ts0 intialized to ' + str(ts0)

        i = 1 # i is the number of successful seeks
        foundfirst = False
        while True:

            # seek to the next frame
            self.source._seek(float(i)/self.fps)

            # get the current time stamp
            ts = self.source.get_next_video_timestamp()

            # did we seek past the end of the movie?
            if ts is None or num.isnan(ts):
                if foundfirst:
                    # we found one keyframe after start, use start
                    self.keyframe_period = i
                    self.keyframe_period_s = self.keyframe_period / self.fps
                else:
                    # then set keyframe period to be length of entire movie
                    self.keyframe_period = self.n_frames + 1
                    self.keyframe_period_s = self.duration_seconds + self.fps
                    if DEBUG_MOVIES: 'Only keyframe found at start of movie, setting keyframe_period = n_frames + 1 = %d, keyframe_period_s = duration_seconds + fps = %f'%(self.keyframe_period,self.keyframe_period_s)

                #raise ValueError( "Could not compute keyframe period in compressed video" )    
                return

            if ts > ts0:
                if foundfirst:
                    break
                else:
                    foundfirst = True
                    i0 = i
                    ts0 = ts

            i = i + 1

        if DEBUG_MOVIES: print "i = %d, i0 = %d"%(i,i0)
        self.keyframe_period = i - i0
        self.keyframe_period_s = self.keyframe_period / self.fps
        if DEBUG_MOVIES: print "Estimated keyframe period = " + str(self.keyframe_period)

    def get_n_frames( self ):
        return self.n_frames

    def get_width( self ):
        return self.width

    def get_height( self ):
        return self.height

    def seek(self,framenumber):

        ts = float(framenumber) / self.fps + self.start_time
        self.source._seek(ts)
        ts = self.source.get_next_video_timestamp()
        self.prevts = ts - 1. / self.fps
        self.currframe = (ts - self.start_time)*self.fps

        return self.currframe

def write_results_to_avi(movie,tracks,filename,f0=None,f1=None):

    nframes = len(tracks)
    if f0 is None:
        f0 = params.start_frame
    if f1 is None:
        f1 = nframes + params.start_frame - 1

    f0 -= params.start_frame
    f1 -= params.start_frame
    f0 = max(0,min(nframes-1,f0))
    f1 = max(0,min(nframes-1,f1))
    nframes_write = f1-f0+1

    # open the file for output
    outstream = open(filename,'wb')

    # write the header
    write_avi_header(movie,tracks,filename,outstream,f0,f1)

    # get the current location
    movilistloc = outstream.tell()

    # write the frames
    offsets = num.zeros(nframes_write)
    for i in range(f0,f1+1):
        if (i % 100) == 0:
            print 'Frame %d / %d'%(i,nframes_write)

        offsets[i-f0] = write_avi_frame(movie,tracks,i,outstream)

        if params.interactive:
            wx.Yield()
        if params.app_instance is not None and not params.app_instance.alive:
            offsets = offsets[:i-f0+1]
            break

    # get offset relative to movilist
    offsets -= movilistloc + 4

    # write the index
    write_avi_index(movie,offsets,outstream)

    # close
    outstream.close()

def write_avi_index(movie,offsets,outstream):

    idx1size = 8 + 16*len( offsets )
    BYTESPERPIXEL = 3
    bytesperframe = int(movie.get_width()*movie.get_height()*BYTESPERPIXEL)

    write_chunk_header('idx1',int(idx1size),outstream)

    for o in offsets:
        try:
            bin_offset = struct.pack( 'I', int(o) )
        except struct.error:
            traceback.print_exc()
            print "writing index %d"%o
            break
        
        outstream.write(struct.pack('4s','00db'))
        outstream.write(struct.pack('I',16))
        outstream.write( bin_offset )
        outstream.write(struct.pack('I',int(bytesperframe)))

def write_avi_frame(movie,tracks,i,outstream):

    height = movie.get_height()
    width = movie.get_width()
    BYTESPERPIXEL = 3
    bytesperframe = width*height*BYTESPERPIXEL

    if tracks is None:
        return
    if i >= len(tracks):
        return

    # global frame index
    j = params.start_frame + i

    # read in the video frame
    try:
        frame, last_timestamp = movie.get_frame(j)
    except (IndexError,NoMoreFramesException):
        return

    # get the current tracks
    ellipses = tracks[i]

    # get tails
    old_pts = []
    early_frame = int(max(0,i-params.tail_length))
    for j in range(early_frame,i+1):
        #print "j = %d"%j
        dataframe = tracks[j]
        #print "dataframe = " + str(dataframe)
        these_pts = []
        ellplot = []
        for ellipse in dataframe.itervalues():
            if num.isnan(ellipse.center.x) or \
                    num.isnan(ellipse.center.y):
                continue
            these_pts.append( (ellipse.center.x,ellipse.center.y,
                               ellipse.identity) )
            ellplot.append(ellipse)
        old_pts.append(these_pts)

    # draw on image
    bitmap,resize,img_size = annotate_bmp(frame,ellplot,old_pts,
                                          params.ellipse_thickness,
                                          [height,width])
    img = bitmap.ConvertToImage()
    # the image is flipped
    img = img.Mirror(True)
    img = img.GetData()

    # write chunktype
    outstream.write(struct.pack('4s','00db'))
    # write size of frame
    outstream.write(struct.pack('I',bytesperframe))

    # write frame
    offset = outstream.tell()
    outstream.write(img[::-1])
    pad = bytesperframe%2
    if pad == 1:
        outstream.write(struct.pack('B',0))
    return offset

def write_avi_header(movie,tracks,filename,outstream,f0,f1):

    # movie size
    BYTESPERPIXEL = 3
    nframes = f1-f0+1
    width = movie.get_width()
    height = movie.get_height()
    bytesperframe = width*height*BYTESPERPIXEL

    # chunk sizes if 0 frames
    avihsize = 64
    #strnsize = 8 + len(filename) + 1
    strllistsize = 116
    strhsize = 56
    strfsize = 48
    hdrllistsize = avihsize + strllistsize + 12
    movilistsize = 12
    idx1size = 8
    riffsize = hdrllistsize + movilistsize + idx1size
    # add in frames
    movilistsize += nframes * (4+4+bytesperframe+(bytesperframe%2))
    idx1size += nframes * (4*4)
    riffsize +=  nframes * (4+4+bytesperframe + 4*4 + (bytesperframe%2))
    ## add in strnsize
    #addon = strnsize + (strnsize%2)
    #riffsize += addon
    #hdrllistsize += addon
    #strllistsize += addon

    # write the RIFF chunk header
    write_chunk_header('RIFF',riffsize,outstream)
    # write AVI fourcc
    outstream.write(struct.pack('4s','AVI '))
    # write hdrl LIST
    write_list_header('hdrl',hdrllistsize-8,outstream)
    # write avih chunk
    write_chunk_header('avih',avihsize-8,outstream)

    ## write main avi header
    # microseconds per frame
    if hasattr(movie,'frame_delay_us'):
        microsecperframe = movie.frame_delay_us
    elif hasattr(movie.h_mov,'frame_delay_us'):
        microsecperframe = movie.h_mov.frame_delay_us
    else:
        microsecperframe = estimate_frame_delay_us(movie.h_mov)
    outstream.write(struct.pack('I',int(round(microsecperframe))))
    # maximum bytes per second
    framespersec = int( round( 1e6/microsecperframe ) )
    bytespersec = framespersec*bytesperframe
    outstream.write(struct.pack('I',int(num.ceil(bytespersec))))
    # reserved
    outstream.write(struct.pack('I',0))
    # flags
    outstream.write(struct.pack('I',16))
    # number of frames
    outstream.write(struct.pack('I',nframes))
    # initial frame
    outstream.write(struct.pack('I',0))
    # number of streams
    outstream.write(struct.pack('I',1))
    # suggested buffer size
    outstream.write(struct.pack('I',bytesperframe))
    # width
    outstream.write(struct.pack('I',width))
    # height
    outstream.write(struct.pack('I',height))
    # frame rate
    outstream.write(struct.pack('2I',100,100*framespersec))
    # not sure -- start, length
    outstream.write(struct.pack('2I',0,0))

    # strl list
    write_list_header('strl',strllistsize-8,outstream)
    # strh chunk
    write_chunk_header('strh',strhsize-8,outstream)

    ## write stream header
    # FCC type
    outstream.write(struct.pack('4s','vids'))
    # FCC handler -- 'DIB '
    #outstream.write(struct.pack('I',0))
    outstream.write( struct.pack( '4s', 'DIB ' ) )
    # Flags
    outstream.write(struct.pack('I',0))
    # Reserved
    outstream.write(struct.pack('I',0))
    # Initial Frame
    outstream.write(struct.pack('I',0))
    # Frame rate
    outstream.write(struct.pack('2I',100,100*framespersec))
    # not sure -- start, length
    outstream.write(struct.pack('2I',0,0))
    # suggested buffer size
    outstream.write(struct.pack('I',bytesperframe))
    # quality
    outstream.write(struct.pack('I',7500))
    # not sure -- sample size
    outstream.write(struct.pack('I',0))

    # Write strf chunk
    write_chunk_header('strf',strfsize-8,outstream)

    ## Write bitmap header
    # Size
    outstream.write(struct.pack('I',40))
    # width
    outstream.write(struct.pack('I',width))
    # height
    outstream.write(struct.pack('I',height))
    # planes
    outstream.write(struct.pack('H',1))
    # bits per pixel
    outstream.write(struct.pack('H',24))
    # FourCC: DIBS
    outstream.write(struct.pack('I',0))
    # image size
    outstream.write(struct.pack('I',bytesperframe))
    # not sure
    outstream.write(struct.pack('4I',0,0,0,0))

    ## Write stream name chunk and data
    #write_chunk_header('strn',strnsize-8,outstream)
    #outstream.write(filename)
    #outstream.write(struct.pack('B',0))
    #if (len(filename)%2) == 1:
    #    outstream.write(struct.pack('B',0))

    # movi list
    write_list_header('movi',movilistsize,outstream)

def write_chunk_header(chunktype,chunksize,outstream):
    try:
        outstream.write(struct.pack('4sI',chunktype,chunksize))
    except struct.error, details:
        traceback.print_exc()
        print "writing '%s' with size %d"%(chunktype,chunksize)
        outstream.write(struct.pack('4sI',chunktype,0))
        

def write_list_header(listtype,listsize,outstream):
    try:
        outstream.write(struct.pack('4sI4s','LIST',listsize,listtype))
    except struct.error, details:
        traceback.print_exc()
        print "writing '%s' with size %d"%(listtype,listsize)
        outstream.write(struct.pack('4sI4s','LIST',0,listtype))


def estimate_frame_delay_us(mov):

    if not hasattr(mov,'chunk_start'):
        return 0

    # go to beginning of first frame
    if mov.issbfmf:
        return .05*1e6
    else:
        mov.file.seek(mov.chunk_start)
        # read the first timestamp
        stamp0 = mov.get_next_timestamp()
        # go to the last frame
        mov.file.seek(mov.chunk_start+mov.bytes_per_chunk*(mov.n_frames-1))
        # read the last timestamp
        stamp1 = mov.get_next_timestamp()


        frame_delay_us = float(stamp1-stamp0)/float(mov.n_frames-1)*1e6
        return frame_delay_us




import os

codedir = os.path.dirname(__file__)
(head,tail) = os.path.split(codedir)
if tail == 'library.zip': # Windows?
    codedir = head
elif tail == 'site-packages.zip': # Mac app
    codedir = head

